import operator
import re
from datetime import date, datetime, timedelta

from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Layout, Submit
from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.template.loader import render_to_string
from django.utils.timezone import make_aware
from django.utils.translation import gettext as _
from django_select2.forms import Select2MultipleWidget
from dynamic_preferences.forms import PreferenceForm
from guardian.shortcuts import assign_perm, get_objects_for_user, get_users_with_perms, remove_perm
from recurrence.forms import RecurrenceField

from ephios.core.dynamic_preferences_registry import event_type_preference_registry
from ephios.core.models import Event, EventType, LocalParticipation, Shift, UserProfile
from ephios.core.signup.methods import enabled_signup_methods
from ephios.core.widgets import MultiUserProfileWidget
from ephios.extra.crispy import AbortLink
from ephios.extra.permissions import get_groups_with_perms
from ephios.extra.widgets import ColorInput, CustomDateInput, CustomTimeInput
from ephios.modellogging.log import add_log_recorder, update_log
from ephios.modellogging.recorders import (
    DerivedFieldsLogRecorder,
    InstanceActionType,
    PermissionLogRecorder,
)


class EventForm(forms.ModelForm):
    visible_for = forms.ModelMultipleChoiceField(
        queryset=Group.objects.none(),
        label=_("Visible for"),
        help_text=_(
            "Select groups which the event shall be visible for. Regardless, the event will be visible for users that already signed up."
        ),
        widget=Select2MultipleWidget,
        required=False,
    )
    responsible_users = forms.ModelMultipleChoiceField(
        queryset=UserProfile.objects.all(),
        required=False,
        label=_("Responsible persons"),
        widget=MultiUserProfileWidget,
    )
    responsible_groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.all(),
        required=False,
        label=_("Responsible groups"),
        widget=Select2MultipleWidget,
    )

    class Meta:
        model = Event
        fields = ["title", "description", "location"]

    def __init__(self, **kwargs):
        user = kwargs.pop("user")
        can_publish_for_groups = get_objects_for_user(user, "publish_event_for_group", klass=Group)

        if (event := kwargs.get("instance", None)) is not None:
            self.eventtype = event.type
            responsible_users = get_users_with_perms(
                event, only_with_perms_in=["change_event"], with_group_users=False
            )
            responsible_groups = get_groups_with_perms(event, only_with_perms_in=["change_event"])
            visible_for = get_groups_with_perms(event, only_with_perms_in=["view_event"]).exclude(
                id__in=responsible_groups
            )

            self.locked_visible_for_groups = set(visible_for.exclude(id__in=can_publish_for_groups))
            kwargs["initial"] = {
                "visible_for": visible_for.filter(id__in=can_publish_for_groups),
                "responsible_users": responsible_users,
                "responsible_groups": responsible_groups,
                **kwargs.get("initial", {}),
            }
        else:
            self.eventtype = kwargs.pop("eventtype")
            kwargs["initial"] = {
                "responsible_users": self.eventtype.preferences.get("responsible_users")
                or get_user_model().objects.filter(pk=user.pk),
                "responsible_groups": self.eventtype.preferences.get("responsible_groups"),
                "visible_for": self.eventtype.preferences.get("visible_for")
                or get_objects_for_user(user, "publish_event_for_group", klass=Group),
            }
            self.locked_visible_for_groups = set()

        super().__init__(**kwargs)

        self.fields["visible_for"].queryset = can_publish_for_groups
        self.fields["visible_for"].disabled = not can_publish_for_groups
        if self.locked_visible_for_groups:
            self.fields["visible_for"].help_text = _(
                "Select groups which the event shall be visible for. "
                "This event is also visible for <b>{groups}</b>, "
                "but you don't have the permission to change visibility "
                "for those groups."
            ).format(groups=", ".join(group.name for group in self.locked_visible_for_groups))

    def save(self, commit=True):
        self.instance.type = self.eventtype
        event: Event = super().save(commit=commit)

        add_log_recorder(event, PermissionLogRecorder("view_event", _("Visible for")))
        add_log_recorder(event, PermissionLogRecorder("change_event", _("Responsibles")))

        # delete existing permissions
        # (better implement https://github.com/django-guardian/django-guardian/issues/654)
        for group in get_groups_with_perms(
            event, only_with_perms_in=["view_event", "change_event"]
        ):
            remove_perm("view_event", group, event)
            remove_perm("change_event", group, event)
        for user in get_users_with_perms(event, only_with_perms_in=["view_event", "change_event"]):
            remove_perm("view_event", user, event)
            remove_perm("change_event", user, event)

        # assign designated permissions
        assign_perm(
            "view_event",
            Group.objects.filter(
                Q(id__in=self.cleaned_data["visible_for"])
                | Q(id__in=self.cleaned_data["responsible_groups"])
                | Q(id__in=(g.id for g in self.locked_visible_for_groups))
            ),
            event,
        )
        assign_perm("change_event", self.cleaned_data["responsible_groups"], event)
        assign_perm("change_event", self.cleaned_data["responsible_users"], event)

        # Assign view_event to responsible users  and to non-responsible users
        # that already have some sort of participation for the event
        # (-> they saw and interacted with it)
        # We can't just do users that aren't included by group permissions,
        # as they might get removed from that group.
        assign_perm(
            "view_event",
            UserProfile.objects.filter(
                Q(pk__in=self.cleaned_data["responsible_users"])
                | Q(
                    pk__in=LocalParticipation.objects.filter(
                        shift_id__in=event.shifts.all()
                    ).values_list("user", flat=True)
                )
            ),
            event,
        )

        update_log(event, InstanceActionType.CHANGE)
        return event


class ShiftForm(forms.ModelForm):
    date = forms.DateField(widget=CustomDateInput, label=_("Date"))
    meeting_time = forms.TimeField(widget=CustomTimeInput, label=_("Meeting time"))
    start_time = forms.TimeField(widget=CustomTimeInput, label=_("Start time"))
    end_time = forms.TimeField(widget=CustomTimeInput, label=_("End time"))

    field_order = ["date", "meeting_time", "start_time", "end_time", "signup_method_slug"]

    class Meta:
        model = Shift
        fields = ["meeting_time", "start_time", "end_time", "signup_method_slug"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signup_methods = list(enabled_signup_methods())
        if self.instance and (method_slug := self.instance.signup_method_slug):
            if method_slug not in map(operator.attrgetter("slug"), signup_methods):
                signup_methods.append(self.instance.signup_method)
        self.fields["signup_method_slug"].widget = forms.Select(
            choices=((method.slug, method.verbose_name) for method in signup_methods)
        )
        # this recorder may cause db queries, so it's added on Shift init, but here in the form
        # pylint: disable=undefined-variable
        add_log_recorder(
            self.instance,
            DerivedFieldsLogRecorder(
                lambda shift: method.get_signup_info() if (method := shift.signup_method) else {}
            ),
        )

    def clean(self):
        cleaned_data = super().clean()
        if {"date", "meeting_time", "start_time", "end_time"} <= set(cleaned_data.keys()):
            cleaned_data["meeting_time"] = make_aware(
                datetime.combine(cleaned_data["date"], cleaned_data["meeting_time"])
            )
            cleaned_data["start_time"] = make_aware(
                datetime.combine(cleaned_data["date"], cleaned_data["start_time"])
            )
            cleaned_data["end_time"] = make_aware(
                datetime.combine(self.cleaned_data["date"], cleaned_data["end_time"])
            )
            if self.cleaned_data["end_time"] <= self.cleaned_data["start_time"]:
                cleaned_data["end_time"] = cleaned_data["end_time"] + timedelta(days=1)
            if not cleaned_data["meeting_time"] <= cleaned_data["start_time"]:
                raise ValidationError(_("Meeting time must not be after start time!"))
        return cleaned_data


class EventDuplicationForm(forms.Form):
    start_date = forms.DateField(
        widget=CustomDateInput,
        initial=date.today(),
        help_text=_(
            "This date will be used as the start date for recurring events that you create below, e.g. daily events will be created from this date onwards."
        ),
        label=_("Start date"),
    )
    recurrence = RecurrenceField(required=False)


class EventTypeForm(forms.ModelForm):
    class Meta:
        model = EventType
        fields = ["title", "color"]
        widgets = {"color": ColorInput()}

    def clean_color(self):
        regex = re.compile(r"#[a-fA-F\d]{6}")
        if not regex.match(self.cleaned_data["color"]):
            raise ValidationError(_("You need to enter a valid color"))
        return self.cleaned_data["color"]


class EventTypePreferenceForm(PreferenceForm):
    registry = event_type_preference_registry


class BaseEventPluginFormMixin:
    @property
    def heading(self):
        raise NotImplementedError

    def render(self):
        try:
            self.helper.form_tag = False
        except AttributeError:
            self.helper = FormHelper(self)
            self.helper.form_tag = False
        return render_to_string("core/fragments/event_plugin_form.html", context={"form": self})

    def is_function_active(self):
        """
        When building forms for additional features, return whether that feature is enabled for the forms event instance.
        With the default template, if this is True, the collapse is expanded on page load.
        """
        return False


class EventNotificationForm(forms.Form):
    NEW_EVENT = "new"
    REMINDER = "remind"
    PARTICIPANTS = "participants"
    action = forms.ChoiceField(
        choices=[
            (NEW_EVENT, _("Send notification about new event to everyone")),
            (REMINDER, _("Send reminder to everyone that is not participating")),
            (PARTICIPANTS, _("Send a message to all participants")),
        ],
        widget=forms.RadioSelect,
        label=False,
    )
    mail_content = forms.CharField(required=False, widget=forms.Textarea, label=_("Mail content"))

    def __init__(self, *args, **kwargs):
        self.event = kwargs.pop("event")
        super().__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.layout = Layout(
            Field("action"),
            Field("mail_content"),
            FormActions(
                Submit("submit", _("Send"), css_class="float-end"),
                AbortLink(href=self.event.get_absolute_url()),
            ),
        )

    def clean(self):
        if (
            self.cleaned_data.get("action") == self.PARTICIPANTS
            and not self.cleaned_data["mail_content"]
        ):
            raise ValidationError(_("You cannot send an empty mail."))
        return super().clean()



# Generated by Django 3.1.6 on 2021-02-07 12:21

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import ephios.extra.json


class Migration(migrations.Migration):

    replaces = [
        ("event_management", "0001_squashed_0009_auto_20200921_2155"),
        ("event_management", "0002_auto_20200927_2157"),
        ("event_management", "0003_auto_20201022_1620"),
        ("event_management", "0004_abstractparticipation_data"),
        ("event_management", "0005_eventtypepreference"),
        ("event_management", "0006_auto_20210204_2257"),
        ("event_management", "0007_auto_20210207_1309"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
        ("core", "0001_squashed_0008_auto_20210207_1309"),
    ]

    operations = [
        migrations.CreateModel(
            name="EventType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("title", models.CharField(max_length=254, verbose_name="title")),
                (
                    "can_grant_qualification",
                    models.BooleanField(verbose_name="can grant qualification"),
                ),
            ],
            options={
                "verbose_name": "event type",
                "verbose_name_plural": "event types",
            },
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("title", models.CharField(max_length=254, verbose_name="title")),
                (
                    "description",
                    models.TextField(blank=True, null=True, verbose_name="description"),
                ),
                ("location", models.CharField(max_length=254, verbose_name="location")),
                (
                    "type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eventtype",
                        verbose_name="event type",
                    ),
                ),
                ("active", models.BooleanField(default=False)),
                (
                    "mail_updates",
                    models.BooleanField(default=True, verbose_name="send updates via mail"),
                ),
            ],
            options={
                "verbose_name": "event",
                "verbose_name_plural": "events",
                "permissions": [("view_past_event", "Can view past events")],
            },
        ),
        migrations.CreateModel(
            name="Shift",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("meeting_time", models.DateTimeField(verbose_name="meeting time")),
                ("start_time", models.DateTimeField(verbose_name="start time")),
                ("end_time", models.DateTimeField(verbose_name="end time")),
                ("signup_method_slug", models.SlugField(verbose_name="signup method")),
                (
                    "signup_configuration",
                    models.JSONField(
                        decoder=ephios.extra.json.CustomJSONDecoder,
                        default=dict,
                        encoder=ephios.extra.json.CustomJSONEncoder,
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="shifts",
                        to="event",
                        verbose_name="shifts",
                    ),
                ),
            ],
            options={
                "verbose_name": "shift",
                "verbose_name_plural": "shifts",
            },
        ),
        migrations.CreateModel(
            name="AbstractParticipation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "state",
                    models.IntegerField(
                        choices=[
                            (0, "requested"),
                            (1, "confirmed"),
                            (2, "declined by user"),
                            (3, "rejected by responsible"),
                        ],
                        default=0,
                        verbose_name="state",
                    ),
                ),
                (
                    "shift",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="shift",
                        verbose_name="shift",
                    ),
                ),
                (
                    "polymorphic_ctype",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="polymorphic_event_management.abstractparticipation_set+",
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "base_manager_name": "objects",
            },
        ),
        migrations.CreateModel(
            name="LocalParticipation",
            fields=[
                (
                    "abstractparticipation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="abstractparticipation",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
                    ),
                ),
            ],
            options={
                "base_manager_name": "objects",
            },
            bases=("core.abstractparticipation",),
        ),
        migrations.AlterModelOptions(
            name="shift",
            options={
                "ordering": ("meeting_time", "start_time", "id"),
                "verbose_name": "shift",
                "verbose_name_plural": "shifts",
            },
        ),
        migrations.AddField(
            model_name="abstractparticipation",
            name="data",
            field=models.JSONField(default=dict),
        ),
        migrations.CreateModel(
            name="EventTypePreference",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        blank=True,
                        db_index=True,
                        default=None,
                        max_length=150,
                        null=True,
                        verbose_name="Section Name",
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=150, verbose_name="Name")),
                ("raw_value", models.TextField(blank=True, null=True, verbose_name="Raw Value")),
                (
                    "instance",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="eventtype"),
                ),
            ],
        ),
        migrations.AlterField(
            model_name="abstractparticipation",
            name="state",
            field=models.IntegerField(
                choices=[
                    (0, "requested"),
                    (1, "confirmed"),
                    (2, "declined by user"),
                    (3, "rejected by responsible"),
                    (4, "getting dispatched"),
                ],
                verbose_name="state",
            ),
        ),
        migrations.AlterModelOptions(
            name="abstractparticipation",
            options={},
        ),
        migrations.AlterModelOptions(
            name="localparticipation",
            options={},
        ),
        migrations.AlterModelTable(
            name="abstractparticipation",
            table="abstractparticipation",
        ),
        migrations.AlterModelTable(
            name="event",
            table="event",
        ),
        migrations.AlterModelTable(
            name="eventtype",
            table="eventtype",
        ),
        migrations.AlterModelTable(
            name="eventtypepreference",
            table="eventtypepreference",
        ),
        migrations.AlterModelTable(
            name="localparticipation",
            table="localparticipation",
        ),
        migrations.AlterModelTable(
            name="shift",
            table="shift",
        ),
    ]



import collections
import operator
from functools import reduce

from django import template
from django.utils import timezone
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from ephios.core.models import AbstractParticipation, EventType, LocalParticipation, UserProfile
from ephios.core.signals import register_event_bulk_action
from ephios.core.views.signup import request_to_participant
from ephios.extra.colors import get_eventtype_color_style

register = template.Library()


@register.filter(name="reverse_signup_action")
def reverse_signup_action(request, shift):
    return request_to_participant(request).reverse_signup_action(shift)


@register.filter(name="participation_from_request")
def participation_from_request(request, shift):
    return request_to_participant(request).participation_for(shift)


@register.filter(name="participation_css_style")
def participation_css_style(participation):
    if participation:
        return {
            AbstractParticipation.States.USER_DECLINED: "danger",
            AbstractParticipation.States.RESPONSIBLE_REJECTED: "danger",
            AbstractParticipation.States.REQUESTED: "warning",
            AbstractParticipation.States.CONFIRMED: "success",
            AbstractParticipation.States.GETTING_DISPATCHED: "(invalid)",
        }[participation.state]
    return "(invalid)"


@register.filter(name="participation_mannequin_style")
def participation_mannequin_style(participation):
    if participation:
        return {
            AbstractParticipation.States.USER_DECLINED: "denied",
            AbstractParticipation.States.RESPONSIBLE_REJECTED: "denied",
            AbstractParticipation.States.REQUESTED: "requested",
            AbstractParticipation.States.CONFIRMED: "confirmed",
            AbstractParticipation.States.GETTING_DISPATCHED: "neutral",
        }[participation.state]
    return "neutral"


@register.simple_tag
def setvar(value=None):
    """
    as an alternative to using with blocks and template, use
    `{% setvar value|filter:param as var %}` to save a value to a variable in a template.
    The builtin `firstof` doesn't work, because it converts values to strings.
    From here: https://pythoncircle.com/post/701/how-to-set-a-variable-in-django-template/
    """
    return value


@register.filter(name="can_sign_up")
def can_sign_up(request, shift):
    return shift.signup_method.can_sign_up(request_to_participant(request))


@register.filter(name="can_customize_signup")
def can_customize_signup(request, shift):
    return shift.signup_method.can_customize_signup(request_to_participant(request))


@register.filter(name="render_shift_state")
def render_shift_state(request, shift):
    return shift.signup_method.render_shift_state(request)


@register.filter(name="signup_errors")
def signup_errors(request, shift):
    return set(
        shift.signup_method.get_signup_errors(request_to_participant(request))
        + shift.signup_method.get_decline_errors(request_to_participant(request))
    )


@register.filter(name="can_decline")
def can_decline(request, shift):
    return shift.signup_method.can_decline(request_to_participant(request))


@register.filter(name="confirmed_participations")
def confirmed_participations(user: UserProfile):
    # can't use order_by, as postgres gets confused when using the
    # `start_time` coalesce, the shift select related and order_by on start_time at the same time.
    return sorted(
        user.participations.filter(state=AbstractParticipation.States.CONFIRMED)
        .filter(end_time__gte=timezone.now())
        .select_related("shift", "shift__event"),
        key=operator.attrgetter("start_time", "end_time"),
    )


@register.filter(name="event_signup_state_counts")
def event_signup_state_counts(event, user):
    """
    Return a counter counting states for a users participations for shifts of an event.
    Uses None as key for no participation info.
    """
    counter = collections.Counter()
    for shift in event.shifts.all():
        for participation in shift.participations.all():
            if isinstance(participation, LocalParticipation) and participation.user_id == user.id:
                counter[participation.state] += 1
                break
    return counter


@register.simple_tag(name="event_bulk_actions")
def event_bulk_actions():
    html = ""
    for _, actions in register_event_bulk_action.send(None):
        html += "".join(
            [
                f"<button class='btn btn-secondary me-1' type='submit' name='{action['url']}' formaction='{action['url']}'><span class='fa {action['icon']}'></span> {action['label']}</button>"
                for action in actions
            ]
        )
    return format_html(html)


@register.simple_tag(name="eventtype_colors")
def eventtype_colors():
    return mark_safe(
        reduce(
            lambda css, eventtype: css + get_eventtype_color_style(eventtype),
            EventType.objects.all(),
            "",
        )
    )


@register.filter(name="color_css")
def eventtype_color_css(eventtype):
    return get_eventtype_color_style(eventtype)



from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import JsonResponse
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView
from django.views.generic.base import View
from django.views.generic.detail import SingleObjectMixin

from ephios.core.consequences import ConsequenceError, editable_consequences
from ephios.core.forms.users import WorkingHourRequestForm


class ConsequenceUpdateView(LoginRequiredMixin, SingleObjectMixin, View):
    def get_queryset(self):
        return editable_consequences(self.request.user)

    def post(self, request, *args, **kwargs):
        consequence = self.get_object()
        fail_reason = None
        if request.POST["action"] == "deny":
            consequence.deny(request.user)
        elif request.POST["action"] == "confirm":
            try:
                consequence.confirm(request.user)
            except ConsequenceError as e:
                fail_reason = str(e)
        return JsonResponse(
            {
                "state": consequence.state,
                "fail_reason": fail_reason,
            }
        )


class WorkingHourRequestView(LoginRequiredMixin, FormView):
    form_class = WorkingHourRequestForm
    template_name = "core/workinghour_request.html"

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["request"] = self.request
        return kwargs

    def form_valid(self, form):
        form.create_consequence()
        messages.success(self.request, _("Your request has been submitted."))
        return redirect(reverse("core:profile"))



from django.contrib import messages
from django.contrib.messages.views import SuccessMessageMixin
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import CreateView, DeleteView, ListView, TemplateView
from django.views.generic.detail import SingleObjectMixin
from dynamic_preferences.forms import preference_form_builder

from ephios.core.forms.events import EventTypeForm, EventTypePreferenceForm
from ephios.core.models import EventType
from ephios.core.views.settings import SettingsViewMixin
from ephios.extra.mixins import CustomPermissionRequiredMixin


class EventTypeUpdateView(
    CustomPermissionRequiredMixin, SettingsViewMixin, TemplateView, SingleObjectMixin
):
    template_name = "core/eventtype_form.html"
    permission_required = "core.change_eventtype"
    model = EventType

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().dispatch(request, *args, **kwargs)

    def get_preference_form(self):
        form_class = preference_form_builder(EventTypePreferenceForm, instance=self.object)
        return form_class(self.request.POST or None)

    def get_context_data(self, **kwargs):
        kwargs.setdefault("form", EventTypeForm(self.request.POST or None, instance=self.object))
        kwargs.setdefault("preference_form", self.get_preference_form())
        return super().get_context_data(**kwargs)

    def post(self, *args, **kwargs):
        form = EventTypeForm(self.request.POST or None, instance=self.object)
        preference_form = self.get_preference_form()

        if all([form.is_valid(), preference_form.is_valid()]):
            event_type = form.save()
            preference_form.update_preferences()
            messages.success(
                self.request, _("The event type {type} has been saved.").format(type=event_type)
            )
            return redirect(reverse("core:settings_eventtype_list"))

        return self.render_to_response(
            self.get_context_data(form=form, preference_form=preference_form)
        )


class EventTypeListView(CustomPermissionRequiredMixin, SettingsViewMixin, ListView):
    permission_required = "core.view_eventtype"
    accept_object_perms = False
    model = EventType


class EventTypeDeleteView(CustomPermissionRequiredMixin, SettingsViewMixin, DeleteView):
    permission_required = "core.delete_eventtype"
    model = EventType

    def get_success_url(self):
        messages.info(
            self.request, _("Eventtype {type} succesfully deleted.").format(type=self.object)
        )
        return reverse("core:settings_eventtype_list")


class EventTypeCreateView(
    CustomPermissionRequiredMixin, SettingsViewMixin, SuccessMessageMixin, CreateView
):
    permission_required = "core.add_eventtype"
    accept_object_perms = False
    template_name = "core/eventtype_form.html"
    model = EventType
    fields = ["title"]
    success_message = _(
        "Event type was created. More settings for this event type can be managed below."
    )

    def get_success_url(self):
        return reverse("core:settings_eventtype_edit", kwargs=dict(pk=self.object.pk))



from django.contrib import messages
from django.contrib.auth.views import redirect_to_login
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import CreateView, DeleteView, DetailView, ListView, UpdateView

from ephios.core.views.settings import SettingsViewMixin
from ephios.extra.mixins import StaffRequiredMixin
from ephios.plugins.pages.models import Page


class PageListView(StaffRequiredMixin, SettingsViewMixin, ListView):
    model = Page


class PageView(DetailView):
    model = Page

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.object = self.get_object()

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_authenticated and not self.object.publicly_visible:
            return redirect_to_login(self.request.get_full_path())
        return super().dispatch(request, *args, **kwargs)


class PageCreateView(StaffRequiredMixin, SettingsViewMixin, CreateView):
    model = Page
    fields = ["title", "content", "slug", "show_in_footer", "publicly_visible"]

    def get_success_url(self):
        messages.success(self.request, _("Page saved successfully."))
        return reverse("pages:settings_page_list")


class PageUpdateView(StaffRequiredMixin, SettingsViewMixin, UpdateView):
    model = Page
    fields = ["title", "content", "slug", "show_in_footer", "publicly_visible"]

    def get_success_url(self):
        messages.success(self.request, _("Page saved successfully."))
        return reverse("pages:settings_page_list")


class PageDeleteView(StaffRequiredMixin, SettingsViewMixin, DeleteView):
    model = Page

    def get_success_url(self):
        messages.info(self.request, _("Page deleted successfully."))
        return reverse("pages:settings_page_list")



try:
    import importlib_metadata  # importlib is broken on python3.8, using backport
except ImportError:
    import importlib.metadata as importlib_metadata

from django.conf import settings
from django.utils.translation import get_language

from ephios.core.models import AbstractParticipation
from ephios.core.signals import footer_link

# suggested in https://github.com/python-poetry/poetry/issues/273
EPHIOS_VERSION = "v" + importlib_metadata.version("ephios")


def ephios_base_context(request):
    footer = {}
    for _, result in footer_link.send(None, request=request):
        for label, url in result.items():
            footer[label] = url

    return {
        "ParticipationStates": AbstractParticipation.States,
        "footer": footer,
        "LANGUAGE_CODE": get_language(),
        "ephios_version": EPHIOS_VERSION,
        "SITE_URL": settings.GET_SITE_URL(),
        "PWA_APP_ICONS": settings.PWA_APP_ICONS,
    }



from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Fieldset, Layout, Submit
from django import forms
from django.contrib.auth.models import Group
from django.db.models import Q
from django.forms import (
    CharField,
    CheckboxSelectMultiple,
    DateField,
    DecimalField,
    Form,
    ModelForm,
    ModelMultipleChoiceField,
    MultipleChoiceField,
    inlineformset_factory,
)
from django.urls import reverse
from django.utils.translation import gettext as _
from django_select2.forms import Select2MultipleWidget, Select2Widget
from guardian.shortcuts import assign_perm, get_objects_for_group, remove_perm

from ephios.core.consequences import WorkingHoursConsequenceHandler
from ephios.core.models import QualificationGrant, UserProfile
from ephios.core.services.notifications.backends import enabled_notification_backends
from ephios.core.services.notifications.types import enabled_notification_types
from ephios.core.widgets import MultiUserProfileWidget
from ephios.extra.crispy import AbortLink
from ephios.extra.permissions import PermissionField, PermissionFormMixin
from ephios.extra.widgets import CustomDateInput
from ephios.modellogging.log import add_log_recorder
from ephios.modellogging.recorders import DerivedFieldsLogRecorder

MANAGEMENT_PERMISSIONS = [
    "auth.add_group",
    "auth.change_group",
    "auth.delete_group",
    "auth.view_group",
    "core.add_userprofile",
    "core.change_userprofile",
    "core.delete_userprofile",
    "core.view_userprofile",
    "core.view_event",
    "core.add_event",
    "core.change_event",
    "core.delete_event",
    "core.view_eventtype",
    "core.add_eventtype",
    "core.change_eventtype",
    "core.delete_eventtype",
    "core.view_qualification",
    "core.add_qualification",
    "core.change_qualification",
    "core.delete_qualification",
    "modellogging.view_logentry",
]


def get_group_permission_log_fields(group):
    # This lives here because it is closely related to the fields on GroupForm below
    if not group.pk:
        return {}
    perms = set(group.permissions.values_list("codename", flat=True))

    return {
        _("Can view past events"): "view_past_event" in perms,
        _("Can add events"): "add_event" in perms,
        _("Can edit users"): "change_userprofile" in perms,
        _("Can manage ephios"): "change_group" in perms,
        # force evaluation of querysets
        _("Can publish events for groups"): set(
            get_objects_for_group(group, "publish_event_for_group", klass=Group)
        ),
        _("Can decide working hours for groups"): set(
            get_objects_for_group(group, "decide_workinghours_for_group", klass=Group)
        ),
    }


class GroupForm(PermissionFormMixin, ModelForm):
    can_view_past_event = PermissionField(
        label=_("Can view past events"), permissions=["core.view_past_event"], required=False
    )

    is_planning_group = PermissionField(
        label=_("Can add events"),
        permissions=["core.add_event", "core.delete_event"],
        required=False,
    )
    publish_event_for_group = ModelMultipleChoiceField(
        label=_("Can publish events for groups"),
        queryset=Group.objects.all(),
        required=False,
        help_text=_("Choose groups that this group can make events visible for."),
        widget=Select2MultipleWidget,
    )
    decide_workinghours_for_group = ModelMultipleChoiceField(
        label=_("Can decide working hours for groups"),
        queryset=Group.objects.all(),
        required=False,
        help_text=_(
            "Choose groups that the group you are currently editing can decide whether to grant working hours for."
        ),
        widget=Select2MultipleWidget,
    )

    is_hr_group = PermissionField(
        label=_("Can edit users"),
        help_text=_(
            "If checked, users in this group can view, add, edit and delete users. They can also manage group memberships for their own groups."
        ),
        permissions=[
            "core.add_userprofile",
            "core.change_userprofile",
            "core.delete_userprofile",
            "core.view_userprofile",
        ],
        required=False,
    )
    is_management_group = PermissionField(
        label=_("Can manage ephios"),
        help_text=_(
            "If checked, users in this group can manage users, groups, all group memberships, eventtypes and qualifications"
        ),
        permissions=MANAGEMENT_PERMISSIONS,
        required=False,
    )

    users = ModelMultipleChoiceField(
        label=_("Users"), queryset=UserProfile.objects.all(), widget=MultiUserProfileWidget
    )

    class Meta:
        model = Group
        fields = ["name"]

    def __init__(self, **kwargs):
        if (group := kwargs.get("instance", None)) is not None:
            kwargs["initial"] = {
                "users": group.user_set.all(),
                "publish_event_for_group": get_objects_for_group(
                    group, "publish_event_for_group", klass=Group
                ),
                "decide_workinghours_for_group": get_objects_for_group(
                    group, "decide_workinghours_for_group", klass=Group
                ),
                **kwargs.get("initial", {}),
            }
            self.permission_target = group
        super().__init__(**kwargs)
        self.helper = FormHelper()
        self.helper.layout = Layout(
            Field("name"),
            Field("users"),
            Field("can_view_past_event"),
            Fieldset(
                _("Management"),
                Field("is_hr_group", title="This permission is included with the management role."),
                "is_management_group",
            ),
            Fieldset(
                _("Planning"),
                Field(
                    "is_planning_group",
                    title="This permission is included with the management role.",
                ),
                Field("publish_event_for_group", wrapper_class="publish-select"),
                "decide_workinghours_for_group",
            ),
            FormActions(
                Submit("submit", _("Save"), css_class="float-end"),
                AbortLink(href=reverse("core:group_list")),
            ),
        )

    def save(self, commit=True):
        add_log_recorder(self.instance, DerivedFieldsLogRecorder(get_group_permission_log_fields))
        group = super().save(commit)

        group.user_set.set(self.cleaned_data["users"])

        remove_perm("publish_event_for_group", group, Group.objects.all())
        if group.permissions.filter(codename="add_event").exists():
            assign_perm(
                "publish_event_for_group", group, self.cleaned_data["publish_event_for_group"]
            )

        if "decide_workinghours_for_group" in self.changed_data:
            remove_perm("decide_workinghours_for_group", group, Group.objects.all())
            assign_perm(
                "decide_workinghours_for_group",
                group,
                self.cleaned_data["decide_workinghours_for_group"],
            )

        group.save()  # logging
        return group


class UserProfileForm(ModelForm):
    groups = ModelMultipleChoiceField(
        label=_("Groups"),
        queryset=Group.objects.all(),
        widget=Select2MultipleWidget,
        required=False,
        disabled=True,  # explicitly enable for users with `change_group` permission
    )

    def __init__(self, *args, **kwargs):
        request = kwargs.pop("request")
        super().__init__(*args, **kwargs)
        self.locked_groups = set()
        if request.user.has_perm("auth.change_group"):
            self.fields["groups"].disabled = False
        elif allowed_groups := request.user.groups:
            self.fields["groups"].disabled = False
            self.fields["groups"].queryset = allowed_groups
            if self.instance.pk:
                self.locked_groups = set(self.instance.groups.exclude(id__in=allowed_groups.all()))
            if self.locked_groups:
                self.fields["groups"].help_text = _(
                    "The user is also member of <b>{groups}</b>, but you are not allowed to manage memberships for those groups."
                ).format(groups=", ".join((group.name for group in self.locked_groups)))

    field_order = [
        "email",
        "first_name",
        "last_name",
        "date_of_birth",
        "phone",
        "groups",
        "is_active",
    ]

    class Meta:
        model = UserProfile
        fields = ["email", "first_name", "last_name", "date_of_birth", "phone", "is_active"]
        widgets = {"date_of_birth": CustomDateInput}
        help_texts = {
            "is_active": _("Inactive users cannot log in and do not get any notifications.")
        }
        labels = {"is_active": _("Active user")}

    def save(self, commit=True):
        userprofile = super().save(commit)
        userprofile.groups.set(
            Group.objects.filter(
                Q(id__in=self.cleaned_data["groups"]) | Q(id__in=(g.id for g in self.locked_groups))
            )
        )
        userprofile.save()
        return userprofile


class QualificationGrantForm(ModelForm):
    model = QualificationGrant

    class Meta:
        fields = ["qualification", "expires"]
        widgets = {"qualification": Select2Widget}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if hasattr(self, "instance") and self.instance.pk:
            # Hide the field and simply display the qualification name in the template
            self.fields["qualification"].disabled = True
            self.fields["qualification"].widget = forms.HiddenInput()
            self.fields["qualification"].title = self.instance.qualification.title


QualificationGrantFormset = inlineformset_factory(
    UserProfile,
    QualificationGrant,
    form=QualificationGrantForm,
    extra=0,
)


class QualificationGrantFormsetHelper(FormHelper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.label_class = "col-md-4"
        self.field_class = "col-md-8"


class WorkingHourRequestForm(Form):
    when = DateField(widget=CustomDateInput, label=_("Date"))
    hours = DecimalField(label=_("Hours of work"), min_value=0.5)
    reason = CharField(label=_("Occasion"))

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)
        self.helper = FormHelper(self)
        self.helper.layout = Layout(
            Field("when"),
            Field("hours"),
            Field("reason"),
            FormActions(
                Submit("submit", _("Send"), css_class="float-end"),
                AbortLink(href=reverse("core:profile")),
            ),
        )

    def create_consequence(self):
        WorkingHoursConsequenceHandler.create(
            user=self.request.user,
            when=self.cleaned_data["when"],
            hours=float(self.cleaned_data["hours"]),
            reason=self.cleaned_data["reason"],
        )


class UserNotificationPreferenceForm(Form):
    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        super().__init__(*args, **kwargs)

        preferences = self.user.preferences["notifications__notifications"]
        for notification_type in enabled_notification_types():
            if notification_type.unsubscribe_allowed:
                self.fields[notification_type.slug] = MultipleChoiceField(
                    label=notification_type.title,
                    choices=[
                        (backend.slug, backend.title) for backend in enabled_notification_backends()
                    ],
                    initial=preferences.get(notification_type.slug, {}),
                    widget=CheckboxSelectMultiple,
                    required=False,
                )

    def update_preferences(self):
        self.user.preferences["notifications__notifications"] = self.cleaned_data



import functools
import operator
from datetime import timedelta
from typing import TYPE_CHECKING

from django.db import models, transaction
from django.db.models import (
    BooleanField,
    CharField,
    DateTimeField,
    ForeignKey,
    IntegerField,
    JSONField,
    Manager,
    Model,
    SlugField,
    TextField,
)
from django.db.models.functions import Coalesce
from django.utils import formats
from django.utils.text import slugify
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _
from dynamic_preferences.models import PerInstancePreferenceModel
from guardian.shortcuts import assign_perm
from polymorphic.managers import PolymorphicManager
from polymorphic.models import PolymorphicModel

from ephios.extra.json import CustomJSONDecoder, CustomJSONEncoder
from ephios.modellogging.log import ModelFieldsLogConfig, register_model_for_logging
from ephios.modellogging.recorders import DerivedFieldsLogRecorder

if TYPE_CHECKING:
    from ephios.core.models import UserProfile
    from ephios.core.signup.methods import BaseSignupMethod, SignupStats
    from ephios.core.signup.participants import AbstractParticipant


class ActiveManager(Manager):
    def get_queryset(self):
        return super().get_queryset().filter(active=True)


class EventType(Model):
    title = CharField(_("title"), max_length=254)
    color = CharField(_("color"), max_length=7, default="#343a40")

    class Meta:
        verbose_name = _("event type")
        verbose_name_plural = _("event types")
        db_table = "eventtype"

    @property
    def color_hex(self):
        return "#FF0"

    def __str__(self):
        return str(self.title)


class Event(Model):
    title = CharField(_("title"), max_length=254)
    description = TextField(_("description"), blank=True, null=True)
    location = CharField(_("location"), max_length=254)
    type = ForeignKey(EventType, on_delete=models.CASCADE, verbose_name=_("event type"))
    active = BooleanField(default=False, verbose_name=_("active"))

    objects = ActiveManager()
    all_objects = Manager()

    class Meta:
        verbose_name = _("event")
        verbose_name_plural = _("events")
        permissions = [("view_past_event", _("Can view past events"))]
        db_table = "event"
        base_manager_name = "all_objects"
        default_manager_name = "objects"

    def get_start_time(self):
        # use shifts.all() in case the shifts have been prefetched
        return getattr(
            self,
            "start_time",
            min(s.start_time for s in self.shifts.all()) if self.shifts.all() else None,
        )

    def get_end_time(self):
        return getattr(
            self,
            "end_time",
            max(s.end_time for s in self.shifts.all()) if self.shifts.all() else None,
        )

    def is_multi_day(self):
        """
        Return whether the event is multi-day. Used to decide whether to show end time or end date.
        """
        start_time = self.get_start_time()
        if start_time is None:
            return False
        # For DLT ambiguity reasons, we cap single-day at 23 hours
        return self.get_end_time() - start_time >= timedelta(hours=23)

    def get_signup_stats(self) -> "SignupStats":
        """Return a SignupStats object aggregated over all shifts of this event, or a default"""
        from ephios.core.signup.methods import SignupStats

        default_for_no_shifts = SignupStats.ZERO

        return functools.reduce(
            operator.add,
            [
                shift.signup_method.get_signup_stats()
                for shift in self.shifts.all()
                if shift.signup_method
            ]
            or [default_for_no_shifts],
        )

    def __str__(self):
        return str(self.title)

    def get_canonical_slug(self):
        return slugify(self.title)

    def get_absolute_url(self):
        from django.urls import reverse

        return reverse("core:event_detail", kwargs=dict(pk=self.id, slug=self.get_canonical_slug()))

    def activate(self):
        if not self.active:
            with transaction.atomic():
                self.active = True
                self.full_clean()
                self.save()


register_model_for_logging(Event, ModelFieldsLogConfig())


class ParticipationManager(PolymorphicManager):
    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                start_time=Coalesce("individual_start_time", "shift__start_time"),
                end_time=Coalesce("individual_end_time", "shift__end_time"),
            )
        )


class DatetimeDisplayMixin:
    """
    Date and time formatting utilities used for start_time and end_time attributes
    in AbstractParticipation and Shift.
    """

    def get_start_time(self):
        return self.start_time

    def get_end_time(self):
        return self.end_time

    def get_date_display(self):
        start_time = localtime(self.get_start_time())
        return f"{formats.date_format(start_time, 'l')}, {formats.date_format(start_time, 'SHORT_DATE_FORMAT')}"

    def get_time_display(self):
        return f"{formats.time_format(localtime(self.get_start_time()))} - {formats.time_format(localtime(self.get_end_time()))}"

    def get_datetime_display(self):
        return f"{self.get_date_display()}, {self.get_time_display()}"


# https://github.com/django-polymorphic/django-polymorphic/issues/229#issuecomment-398434412
def NON_POLYMORPHIC_CASCADE(collector, field, sub_objs, using):
    return models.CASCADE(collector, field, sub_objs.non_polymorphic(), using)


class AbstractParticipation(DatetimeDisplayMixin, PolymorphicModel):
    class States(models.IntegerChoices):
        REQUESTED = 0, _("requested")
        CONFIRMED = 1, _("confirmed")
        USER_DECLINED = 2, _("declined by user")
        RESPONSIBLE_REJECTED = 3, _("rejected by responsible")
        GETTING_DISPATCHED = 4, _("getting dispatched")

        @classmethod
        def labels_dict(cls):
            return dict(zip(cls.values, cls.labels))

    shift = ForeignKey(
        "Shift",
        on_delete=NON_POLYMORPHIC_CASCADE,
        verbose_name=_("shift"),
        related_name="participations",
    )
    state = IntegerField(_("state"), choices=States.choices, default=States.GETTING_DISPATCHED)
    data = models.JSONField(default=dict, verbose_name=_("Signup data"))

    """
    Overwrites shift time. Use `start_time` and `end_time` to get the applicable time (implemented with a custom manager).
    """
    individual_start_time = DateTimeField(_("individual start time"), null=True)
    individual_end_time = DateTimeField(_("individual end time"), null=True)

    # human readable comment
    comment = models.CharField(_("Comment"), max_length=255, blank=True)

    """
    The finished flag is used to make sure the participation_finished signal is only sent out once, even
    if the shift time is changed afterwards.
    """
    finished = models.BooleanField(default=False, verbose_name=_("finished"))

    objects = ParticipationManager()

    def has_customized_signup(self):
        return bool(
            self.individual_start_time
            or self.individual_end_time
            or self.comment
            or self.shift.signup_method.has_customized_signup(self)
        )

    @property
    def hours_value(self):
        td = self.shift.end_time - self.shift.start_time
        return td.total_seconds() / (60 * 60)

    @property
    def participant(self) -> "AbstractParticipant":
        raise NotImplementedError

    class Meta:
        db_table = "abstractparticipation"

    def __str__(self):
        try:
            return f"{self.participant} @ {self.shift}"
        except NotImplementedError:
            return super().__str__()

    def is_in_positive_state(self):
        return self.state in {self.States.CONFIRMED, self.States.REQUESTED}


PARTICIPATION_LOG_CONFIG = ModelFieldsLogConfig(
    unlogged_fields=["id", "data", "abstractparticipation_ptr"],
    attach_to_func=lambda instance: (Event, instance.shift.event_id),
)


class Shift(DatetimeDisplayMixin, Model):
    event = ForeignKey(
        Event, on_delete=models.CASCADE, related_name="shifts", verbose_name=_("event")
    )
    meeting_time = DateTimeField(_("meeting time"))
    start_time = DateTimeField(_("start time"))
    end_time = DateTimeField(_("end time"))
    signup_method_slug = SlugField(_("signup method"))
    signup_configuration = JSONField(
        default=dict, encoder=CustomJSONEncoder, decoder=CustomJSONDecoder
    )

    class Meta:
        verbose_name = _("shift")
        verbose_name_plural = _("shifts")
        ordering = ("meeting_time", "start_time", "id")
        db_table = "shift"

    @property
    def signup_method(self) -> "BaseSignupMethod":
        from ephios.core.signup.methods import signup_method_from_slug

        try:
            event = self.event
        except Event.DoesNotExist:
            event = None
        try:
            return signup_method_from_slug(self.signup_method_slug, self, event=event)
        except ValueError:
            return None

    def get_participants(self, with_state_in=frozenset({AbstractParticipation.States.CONFIRMED})):
        for participation in self.participations.filter(state__in=with_state_in):
            yield participation.participant

    def get_absolute_url(self):
        return f"{self.event.get_absolute_url()}#shift-{self.pk}"

    def __str__(self):
        return f"{self.event.title} ({self.get_datetime_display()})"


class ShiftLogConfig(ModelFieldsLogConfig):
    def __init__(self):
        super().__init__(unlogged_fields=["id", "signup_method_slug", "signup_configuration"])

    def object_to_attach_logentries_to(self, instance):
        return Event, instance.event_id

    def initial_log_recorders(self, instance):
        # pylint: disable=undefined-variable
        yield from super().initial_log_recorders(instance)

        def get_signup_method_name_mapping(shift):
            from ephios.core.signup.methods import installed_signup_methods

            v = None
            for method in installed_signup_methods():
                if method.slug == shift.signup_method_slug:
                    v = str(method.verbose_name)
            return {_("Signup method"): v}

        yield DerivedFieldsLogRecorder(get_signup_method_name_mapping)


register_model_for_logging(Shift, ShiftLogConfig())


class LocalParticipation(AbstractParticipation):
    user: "UserProfile" = ForeignKey(
        "UserProfile",
        on_delete=models.CASCADE,
        verbose_name=_("Participant"),
        related_name="participations",
    )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if (
            not self.user.has_perm("core.view_event", obj=self.shift.event)
            and self.state != self.States.GETTING_DISPATCHED
        ):
            # If dispatched by a responsible, the user should be able to view
            # the event, if not already permitted through its group.
            # Currently, this permission does not get removed automatically.
            assign_perm("core.view_event", user_or_group=self.user, obj=self.shift.event)

    @property
    def participant(self):
        return self.user.as_participant()

    class Meta:
        db_table = "localparticipation"


register_model_for_logging(LocalParticipation, PARTICIPATION_LOG_CONFIG)


class PlaceholderParticipation(AbstractParticipation):
    first_name = CharField(max_length=254)
    last_name = CharField(max_length=254)

    @property
    def participant(self) -> "AbstractParticipant":
        from ephios.core.models.users import Qualification
        from ephios.core.signup.participants import PlaceholderParticipant

        return PlaceholderParticipant(
            first_name=self.first_name,
            last_name=self.last_name,
            qualifications=Qualification.objects.none(),
            email=None,
            date_of_birth=None,
        )

    def __str__(self):
        return f"{self.first_name} {self.last_name} @ {self.shift}"


register_model_for_logging(PlaceholderParticipation, PARTICIPATION_LOG_CONFIG)


class EventTypePreference(PerInstancePreferenceModel):
    instance = ForeignKey(EventType, on_delete=models.CASCADE)

    class Meta:
        db_table = "eventtypepreference"
        app_label = "core"  # https://github.com/agateblue/django-dynamic-preferences/issues/96



import dataclasses
import secrets

from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ephios.core.models import AbstractParticipation, Event, Qualification
from ephios.core.models.events import PARTICIPATION_LOG_CONFIG
from ephios.core.signup.participants import AbstractParticipant
from ephios.modellogging.log import register_model_for_logging


class EventGuestShare(models.Model):
    event = models.OneToOneField(Event, related_name="guest_share", on_delete=models.CASCADE)
    token = models.CharField(max_length=254, default=secrets.token_urlsafe, unique=True)
    active = models.BooleanField(default=False)

    def new_token(self):
        self.token = secrets.token_urlsafe()

    @property
    def url(self):
        return settings.GET_SITE_URL() + reverse(
            "guests:register", kwargs=dict(public_signup_token=self.token, event_id=self.event.id)
        )


class GuestUser(models.Model):
    email = models.EmailField(_("email address"))
    first_name = models.CharField(_("first name"), max_length=254)
    last_name = models.CharField(_("last name"), max_length=254)
    date_of_birth = models.DateField(_("date of birth"))
    phone = models.CharField(_("phone number"), max_length=254, blank=True)
    access_token = models.CharField(max_length=254, default=secrets.token_urlsafe, unique=True)
    qualifications = models.ManyToManyField(Qualification)
    event = models.ForeignKey(Event, on_delete=models.CASCADE)
    created_at = models.DateTimeField(auto_now_add=True)

    def as_participant(self) -> "GuestParticipant":
        return GuestParticipant(
            first_name=self.first_name,
            last_name=self.last_name,
            qualifications=self.qualifications.all(),
            date_of_birth=self.date_of_birth,
            email=self.email,
            guest_user=self,
        )

    def __str__(self):
        return f"{self.first_name} {self.last_name} @ {self.event}"

    class Meta:
        # there might be two people using the same email *sigh*
        unique_together = [["event", "email", "first_name", "last_name"]]


class GuestParticipation(AbstractParticipation):
    guest_user = models.ForeignKey(
        GuestUser, on_delete=models.CASCADE, verbose_name=_("guest participant")
    )

    @property
    def participant(self) -> AbstractParticipant:
        return self.guest_user.as_participant()


register_model_for_logging(GuestParticipation, PARTICIPATION_LOG_CONFIG)


@dataclasses.dataclass(frozen=True)
class GuestParticipant(AbstractParticipant):
    guest_user: GuestUser

    def new_participation(self, shift):
        return GuestParticipation(shift=shift, guest_user=self.guest_user)

    def participation_for(self, shift):
        try:
            return GuestParticipation.objects.get(shift=shift, guest_user=self.guest_user)
        except GuestParticipation.DoesNotExist:
            return None

    def all_participations(self):
        return GuestParticipation.objects.filter(guest_user=self.guest_user)

    def reverse_signup_action(self, shift):
        return reverse(
            "guests:signup_action",
            kwargs=dict(pk=shift.pk, guest_access_token=self.guest_user.access_token),
        )

    def reverse_event_detail(self, event):
        return reverse(
            "guests:event_detail", kwargs=dict(guest_access_token=self.guest_user.access_token)
        )

    @property
    def icon(self):
        return mark_safe(
            f'<span class="fa fa-user-tag" data-bs-toggle="tooltip" data-bs-placement="left" title="{_("Guest")}"></span>'
        )



"""ephios URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.conf import settings
from django.contrib import admin
from django.urls import include, path
from django.views.i18n import JavaScriptCatalog

from ephios.core.plugins import get_all_plugins

urlpatterns = [
    path("admin/", admin.site.urls),
    path("accounts/", include("django.contrib.auth.urls")),
    path("select2/", include("django_select2.urls")),
    path("webpush/", include("webpush.urls")),
    path("", include("ephios.core.urls")),
    path("jsi18n.js", JavaScriptCatalog.as_view(packages=["recurrence"]), name="jsi18n"),
]

# Insert plugin url configs. We can't easily restrict to enabled plugins, as patterns are collected on startup.
for plugin in get_all_plugins():
    try:
        urlpatterns.append(path("", include(plugin.module + ".urls")))
    except ModuleNotFoundError:
        pass

if settings.DEBUG:
    import debug_toolbar

    urlpatterns = [
        path("__debug__/", include(debug_toolbar.urls)),
    ] + urlpatterns



# Generated by Django 3.1.7 on 2021-03-19 16:04

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import ephios.extra.json


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0005_remove_event_mail_updates"),
    ]

    operations = [
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("slug", models.SlugField(max_length=255)),
                ("failed", models.BooleanField(default=False)),
                (
                    "data",
                    models.JSONField(
                        blank=True,
                        decoder=ephios.extra.json.CustomJSONDecoder,
                        default=dict,
                        encoder=ephios.extra.json.CustomJSONEncoder,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="affected user",
                    ),
                ),
            ],
        ),
    ]



import itertools
from enum import Enum
from typing import Callable

from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist
from django.db import models
from django.db.models.signals import m2m_changed
from django.dispatch import Signal, receiver
from django.template.defaultfilters import yesno
from django.utils.translation import gettext_lazy as _
from guardian.shortcuts import get_users_with_perms

from ephios.extra.permissions import get_groups_with_perms

# pylint: disable=protected-access


def capitalize_first(string):
    return string[0].upper() + string[1:]


class InstanceActionType(str, Enum):
    CREATE = "create"
    CHANGE = "change"
    DELETE = "delete"


class BaseLogRecorder:
    """
    A Log Recorder is used to record changes made to a model instance that can later be included in a log entry.
    A recorder follows a strict lifecycle:
        - creation at model instance initialisation or later
        - getting ``attached`` to an instance
        - ``record``ing a change from an instance and the way it exists in the db, or just saving the current state
        - ``is_changed`` is called to find out whether the recorder should be included in the log
        - ``key`` is used to merge multiple log entries into one, should there be multiple ``save`` calls
        - ``serialize`` is called to save the state into json. The ``slug`` is then added to later identify the corresponding Recorder class.
        - ``deserialize`` is called when we want to display a recorder's recordings after some time.
          deserialized recorders will never be serialized again, as e.g. a model field might not exist anymore
        - ``change_statements`` (for change action) and ``value_statements`` (for create and delete action)
           then return dicts to be fed into a statement template.

    Statement templates accept a ``label`` value, an optional ``verb``, and either an ``objects`` list or
    a ``value`` with - optionally - an additional ``old_value``.
    """

    slug = NotImplemented

    def attached(self, instance):
        pass

    def record(self, action: InstanceActionType, instance):
        """
        Record changes for change action and field state for creation and deletion.
        If action is CHANGE,``db_instance`` is set to an instance freshly fetched from the db.
        """

    def is_changed(self):
        """
        Returns whether the values have changed and should be included in the log.
        """
        raise NotImplementedError

    @property
    def key(self):
        """
        Some key to identify the field
        """
        raise NotImplementedError

    def serialize(self, action_type: InstanceActionType):
        """
        dump this change to a dict
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        """
        load this change from a dict.
        """
        raise NotImplementedError

    def change_statements(self):
        """
        Yield human readable statements for reporting about changes.
        """
        raise NotImplementedError

    def value_statements(self):
        """
        Yield human readable statements about the current value.
        Used for reporting about created and deleted instances.
        """
        raise NotImplementedError


class ModelFieldLogRecorder(BaseLogRecorder):
    slug = "modelfield"

    def __init__(self, field):
        self.field = field

    def attached(self, instance):
        # is id for foreign keys
        self.old_value = self.field.value_from_object(instance)

    def record(self, action: InstanceActionType, instance):
        if self.field.one_to_one or self.field.many_to_one:
            # actual object for foreign keys
            try:
                self.new_value = getattr(instance, self.field.name)
            except ObjectDoesNotExist:
                self.new_value = None
        else:
            self.new_value = self.field.value_from_object(instance)

    def is_changed(self):
        if self.field.one_to_one or self.field.many_to_one:
            return self.old_value != getattr(self.new_value, "pk", None)
        return self.old_value != self.new_value

    @property
    def key(self):
        return f"field-{self.field.name}"

    def serialize(self, action_type: InstanceActionType):
        data = {
            "field_name": self.field.name,
            "verbose_name": str(
                getattr(self.field, "verbose_name", capitalize_first(self.field.name))
            ),
            "old_value": self.old_value,
            "new_value": self.new_value,
        }
        if self.field.one_to_one or self.field.many_to_one:
            if not self.is_changed():
                data["old_value"] = self.new_value
            elif self.old_value:
                try:
                    data["old_value"] = self.field.related_model._base_manager.get(
                        pk=self.old_value
                    )
                except self.field.related_model.DoesNotExist:
                    pass
        return data

    def _choice_to_display(self, choice):  # does not support nested choices
        for key, label in self.field.choices:
            if key == choice:
                return label
        return choice

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        try:
            self = cls(model._meta.get_field(data["field_name"]))
        except (FieldDoesNotExist, AttributeError):
            self = cls(None)
            self.label = data["verbose_name"]
        else:
            self.label = getattr(self.field, "verbose_name", capitalize_first(self.field.name))

        self.old_value = data["old_value"]
        self.new_value = data["new_value"]

        if self.field is None:
            pass
        elif self.field.one_to_one or self.field.many_to_one:
            pass
        elif getattr(self.field, "choices", None):
            self.old_value = self._choice_to_display(self.old_value)
            self.new_value = self._choice_to_display(self.new_value)
        elif isinstance(self.field, models.BooleanField):
            self.old_value = yesno(self.old_value)
            self.new_value = yesno(self.new_value)

        return self

    def change_statements(self):
        yield dict(label=self.label, value=self.new_value, old_value=self.old_value)

    def value_statements(self):
        yield dict(label=self.label, value=self.new_value)


class M2MLogRecorder(BaseLogRecorder):
    # pylint: disable=too-many-instance-attributes

    slug = "m2mfield"

    def __init__(self, field, reverse=False, verbose_name=None):
        if reverse and not verbose_name:
            raise ValueError("You must provide a verbose name when reversing m2m relationships.")

        self.field = field
        self.reversed = reverse
        self.verbose_name = verbose_name
        if field and not self.verbose_name:
            self.verbose_name = str(getattr(field, "verbose_name", capitalize_first(field.name)))
        self.added_pks, self.removed_pks = set(), set()

    def attached(self, instance):
        self.model = type(instance)

    @property
    def field_name(self):
        return self.field.remote_field.related_name if self.reversed else self.field.name

    def record(self, action: InstanceActionType, instance):
        if action != InstanceActionType.CHANGE:
            self.current = list(getattr(instance, self.field_name).all())

    def record_clear(self, instance):
        self.added_pks = set()
        self.removed_pks = {obj.pk for obj in getattr(instance, self.field_name).all()}

    def is_changed(self):
        return bool(self.removed_pks) or bool(self.added_pks)

    @property
    def key(self):
        return f"m2m-{self.field.name}"

    def serialize(self, action_type: InstanceActionType):
        related_model = self.field.model if self.reversed else self.field.related_model
        data = {
            "field_name": self.field_name,
            "verbose_name": self.verbose_name,
            "added": related_model._base_manager.filter(pk__in=self.added_pks),
            "removed": related_model._base_manager.filter(pk__in=self.removed_pks),
        }

        if current := getattr(self, "current", None):
            data["current"] = current

        return data

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        try:
            self = cls(model._meta.get_field(data["field_name"]))
        except FieldDoesNotExist:
            self = cls(None)
            self.label = data["verbose_name"]
        else:
            self.label = getattr(self.field, "verbose_name", capitalize_first(self.field.name))
        self.added = data["added"]
        self.removed = data["removed"]
        self.current = data.get("current", [])
        return self

    def change_statements(self):
        for attr, verb in [("added", _("added")), ("removed", _("removed"))]:
            if objects := getattr(self, attr):
                yield dict(label=self.label, verb=verb, objects=objects)

    def value_statements(self):
        if self.current:
            yield dict(label=self.label, objects=self.current)


@receiver(m2m_changed)
def _m2m_changed(
    sender, instance, action, reverse, model, pk_set, **kwargs
):  # pylint: disable=unused-argument
    from ephios.modellogging.log import LOGGED_MODELS, update_log

    if "pre" not in action:
        return

    if type(instance) not in LOGGED_MODELS:
        return

    hit = False
    for recorder in instance._log_recorders:
        if recorder.slug != M2MLogRecorder.slug:
            continue
        if getattr(type(instance), recorder.field_name).through == sender:
            if action == "pre_remove":
                recorder.removed_pks |= set(pk_set)
                hit = True
            elif action == "pre_add":
                recorder.added_pks |= set(pk_set)
                hit = True
            elif action == "pre_clear":
                recorder.record_clear(instance)
                hit = True
        if hit:
            update_log(instance, InstanceActionType.CHANGE)


class PermissionLogRecorder(BaseLogRecorder):
    """
    This recorder records users and groups that have object permissions for the logged instance.
    If the logged model is User or Group, consider using the DerivedFieldsLogRecorder to record Permissions.
    """

    # pylint: disable=too-many-instance-attributes
    slug = "permission-recorder"

    def __init__(self, codename, label):
        self.codename = codename
        self.label = label

    def attached(self, instance):
        self.old_users = set(
            get_users_with_perms(
                instance, with_group_users=False, only_with_perms_in=[self.codename]
            )
        )
        self.old_groups = set(get_groups_with_perms(instance, only_with_perms_in=[self.codename]))

    def record(self, action: InstanceActionType, instance):
        self.new_users = set(
            get_users_with_perms(
                instance, with_group_users=False, only_with_perms_in=[self.codename]
            )
        )
        self.new_groups = set(get_groups_with_perms(instance, only_with_perms_in=[self.codename]))

    def is_changed(self):
        return self.new_groups != self.old_groups or self.new_users != self.old_users

    @property
    def key(self):
        return f"permission-{self.codename}"

    def serialize(self, action_type: InstanceActionType):
        data = {
            "label": str(self.label),
            "codename": self.codename,
        }

        if action_type == InstanceActionType.CHANGE:
            data["added_users"] = self.new_users - self.old_users
            data["removed_users"] = self.old_users - self.new_users
            data["added_groups"] = self.new_groups - self.old_groups
            data["removed_groups"] = self.old_groups - self.new_groups
        else:
            data["users"] = self.new_users
            data["groups"] = self.old_groups

        return data

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        self = cls(data["codename"], data["label"])
        if action_type == InstanceActionType.CHANGE:
            self.added = list(map(str, itertools.chain(data["added_groups"], data["added_users"])))
            self.removed = list(
                map(str, itertools.chain(data["removed_groups"], data["removed_users"]))
            )
        else:
            self.current = list(map(str, itertools.chain(data["groups"], data["users"])))

        return self

    def change_statements(self):
        for attr, verb in [("added", _("added")), ("removed", _("removed"))]:
            if objects := getattr(self, attr):
                yield dict(label=self.label, verb=verb, objects=objects)

    def value_statements(self):
        if hasattr(self, "current"):
            yield dict(label=self.label, objects=self.current)


class DerivedFieldsLogRecorder(BaseLogRecorder):
    """
    This recorder lets you provide a ``derive`` function you can use to derive fields for the log from the instance.
    Return value must be a dict mapping human readable labels to values.
    """

    slug = "derived-fields"

    def __init__(self, derive: Callable):
        self.derive = derive

    def attached(self, instance):
        self.old_dict = self.derive(instance)

    def record(self, action: InstanceActionType, instance):
        self.new_dict = self.derive(instance)

    def is_changed(self):
        return self.old_dict != self.new_dict

    @property
    def key(self):
        return f"derived-{id(self.derive)}"

    def serialize(self, action_type: InstanceActionType):
        return dict(
            changes={
                str(key): [old_value, new_value]
                for key in set(itertools.chain(self.old_dict.keys(), self.new_dict.keys()))
                if (old_value := self.old_dict.get(key)) != (new_value := self.new_dict.get(key))
                or action_type != InstanceActionType.CHANGE
            }
        )

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        self = cls(None)
        self.changes = data["changes"]

        def prettify(value):
            if isinstance(value, bool):
                return yesno(value)
            if isinstance(value, (list, set)):
                return ", ".join(map(str, value))
            return value

        for key in self.changes:
            old, new = self.changes[key]
            self.changes[key] = [prettify(old), prettify(new)]
        return self

    def change_statements(self):
        for label, (old_value, new_value) in self.changes.items():
            yield dict(label=label, value=new_value, old_value=old_value)

    def value_statements(self):
        for label, (__, new_value) in self.changes.items():
            yield dict(label=label, value=new_value)


class FixedMessageLogRecorder(BaseLogRecorder):
    slug = "fixed-message"

    def __init__(self, label, message, show_change_statement=True, show_value_statement=False):
        self.label = str(label)
        self.message = str(message)
        self.show_change_statement = show_change_statement
        self.show_value_statement = show_value_statement

    def is_changed(self):
        return True

    @property
    def key(self):
        return f"fixed-{id(self.label + self.message)}"

    def serialize(self, action_type: InstanceActionType):
        return self.__dict__

    @classmethod
    def deserialize(cls, data, model, action_type: InstanceActionType):
        return cls(**data)

    def change_statements(self):
        if self.show_change_statement:
            yield dict(label=self.label, value=self.message)

    def value_statements(self):
        if self.show_value_statement:
            yield dict(label=self.label, value=self.message)


register_log_recorders = Signal()


@receiver(register_log_recorders)
def _register_inbuilt_recorders(sender, **kwargs):
    return [
        ModelFieldLogRecorder,
        M2MLogRecorder,
        PermissionLogRecorder,
        DerivedFieldsLogRecorder,
        FixedMessageLogRecorder,
    ]


def recorder_types_by_slug(model):
    return {
        recorder.slug: recorder
        for recorder in itertools.chain(
            *(recorders for __, recorders in register_log_recorders.send(model))
        )
    }



# Generated by Django 3.1.7 on 2021-04-03 21:48

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0008_auto_20210403_2348"),
        ("eventautoqualification", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="eventautoqualificationconfiguration",
            options={"verbose_name": "event auto qualification configuration"},
        ),
        migrations.AlterField(
            model_name="eventautoqualificationconfiguration",
            name="mode",
            field=models.IntegerField(
                choices=[(1, "any shift"), (2, "every shift"), (3, "last shift")],
                default=1,
                verbose_name="Required attendance",
            ),
        ),
        migrations.AlterField(
            model_name="eventautoqualificationconfiguration",
            name="qualification",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="core.qualification",
                verbose_name="Qualification",
            ),
        ),
    ]



from django.apps import AppConfig
from dynamic_preferences.registries import preference_models


class UserManagementConfig(AppConfig):
    name = "ephios.core"

    def ready(self):
        from ephios.core.dynamic_preferences_registry import event_type_preference_registry

        from . import signals  # pylint: disable=unused-import

        EventTypePreference = self.get_model("EventTypePreference")
        preference_models.register(EventTypePreference, event_type_preference_registry)



import itertools
import threading
from typing import Dict

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.signals import post_init, post_save, pre_delete, pre_save
from django.dispatch import receiver

from ephios.modellogging.models import LogEntry
from ephios.modellogging.recorders import InstanceActionType, M2MLogRecorder, ModelFieldLogRecorder

# pylint: disable=protected-access


class BaseLogConfig:
    def initial_log_recorders(self, instance):
        """
        Initial log recorders are always added after model __init__.
        Therefore, recorders added here should not do any db operations.
        If you need those, add them manually, e.g. at the beginning of a form's save().
        """
        return []

    def related_logentries(self, instance):
        """
        Return a queryset with all logentries that should be shown with the instance.
        """
        return LogEntry.objects.filter(
            attached_to_object_type=ContentType.objects.get_for_model(type(instance)),
            attached_to_object_id=instance.pk,
        )

    def object_to_attach_logentries_to(self, instance):
        """
        Return a model class and primary key for the object for which the instance's logentry should be shown.
        By default, show it to the object described by the logentry itself.

        Returning the model instance directly might rely on fetching that object from the database,
        which can break bulk loading in some cases, so we don't do that.
        """
        return type(instance), instance.pk


class ModelFieldsLogConfig(BaseLogConfig):
    def __init__(self, unlogged_fields=None, attach_to_func=None, initial_recorders_func=None):
        """
        ``unlogged_fields``: Specify a list of field names so that these fields don't get logged.
        Other fields get logged. Defaults to ['id']

        ``attach_to_func``: Specify a function receiving an instance and returning a tuple
        of model class and object pk to attach log entries to.
        Defaults to the instance's class and its own pk.

        ``initial_recorders_func``: Specify a function receiving an instance and returning
        a list of additional log recorders.
        """

        if unlogged_fields is None:
            unlogged_fields = ["id"]
        self.unlogged_fields = unlogged_fields
        self.attach_to_func = attach_to_func
        self.initial_recorders_func = initial_recorders_func

    def object_to_attach_logentries_to(self, instance):
        if self.attach_to_func:
            return self.attach_to_func(instance)
        return super().object_to_attach_logentries_to(instance)

    def initial_log_recorders(self, instance):
        opts = instance._meta
        for f in itertools.chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
            if not getattr(f, "editable", False):
                continue
            if f.name in self.unlogged_fields:
                continue
            if f.many_to_many:
                yield M2MLogRecorder(f)
            else:
                yield ModelFieldLogRecorder(f)
        if self.initial_recorders_func:
            yield from self.initial_recorders_func(instance)


log_request_store = threading.local()

LOGGED_MODELS: Dict[models.Model, BaseLogConfig] = {}


def register_model_for_logging(model_class, config):
    LOGGED_MODELS[model_class] = config


def add_log_recorder(instance, recorder):
    instance._log_recorders.append(recorder)
    recorder.attached(instance)


def _get_log_data(instance, action_type):
    for recorder in instance._log_recorders:
        recorder.record(action_type, instance)

    data = {
        recorder.key: {
            "slug": recorder.slug,
            "data": recorder.serialize(action_type),
        }
        for recorder in instance._log_recorders
        if recorder.is_changed() or action_type != InstanceActionType.CHANGE
    }

    if not data:
        return {}

    try:
        data["__str__"] = str(instance)
    except ObjectDoesNotExist:
        # sometimes, when deleting objects, they are in an invalid state, so str fails
        if not action_type == InstanceActionType.DELETE:
            raise
        data["__str__"] = None
    return data


def update_log(instance, action_type: InstanceActionType):
    logentry = getattr(instance, "_current_logentry", None)
    log_data = _get_log_data(instance, logentry.action_type if logentry else action_type)
    if not log_data:
        return

    if logentry:
        logentry.data.update(log_data)
    else:
        try:
            user = log_request_store.request.user
            if not user.is_authenticated:
                user = None
            request_id = log_request_store.request_id
        except AttributeError:
            user = None
            request_id = None
        config = LOGGED_MODELS[type(instance)]
        attach_to_model, attached_to_object_id = config.object_to_attach_logentries_to(instance)
        attached_to_object_type = ContentType.objects.get_for_model(attach_to_model)
        logentry = LogEntry(
            content_object=instance,
            attached_to_object_type=attached_to_object_type,
            attached_to_object_id=attached_to_object_id,
            user=user,
            request_id=request_id,
            action_type=action_type,
            data=log_data,
        )
    logentry.save()
    instance._current_logentry = logentry


@receiver(post_init)
def log_post_init(sender, instance, **kwargs):
    if config := LOGGED_MODELS.get(sender):
        instance._log_recorders = list(config.initial_log_recorders(instance))
        for recorder in instance._log_recorders:
            recorder.attached(instance)


@receiver(pre_save)
def log_pre_save(sender, instance, raw, **kwargs):
    if instance.pk and not raw and sender in LOGGED_MODELS:
        update_log(instance, InstanceActionType.CHANGE)


@receiver(post_save)
def log_post_save(sender, instance, created, raw, **kwargs):
    if created and not raw and sender in LOGGED_MODELS:
        update_log(instance, InstanceActionType.CREATE)


@receiver(pre_delete)
def log_pre_delete(sender, instance, **kwargs):
    if sender in LOGGED_MODELS:
        update_log(instance, InstanceActionType.DELETE)



# Generated by Django 3.2.5 on 2021-07-03 13:39

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0009_auto_20210528_2230"),
    ]

    operations = [
        migrations.CreateModel(
            name="PlaceholderParticipation",
            fields=[
                (
                    "abstractparticipation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.abstractparticipation",
                    ),
                ),
                ("first_name", models.CharField(max_length=254)),
                ("last_name", models.CharField(max_length=254)),
            ],
            options={
                "abstract": False,
                "base_manager_name": "objects",
            },
            bases=("core.abstractparticipation",),
        ),
    ]



from calendar import HTMLCalendar, day_abbr
from datetime import date, datetime
from itertools import groupby

from django.utils.formats import date_format
from django.utils.translation import gettext as _


class ShiftCalendar(HTMLCalendar):
    cssclass_month = "table table-fixed"

    def __init__(self, shifts, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.shifts = {
            k: list(v) for (k, v) in groupby(shifts, lambda shift: shift.start_time.date().day)
        }

    def formatmonth(self, theyear, themonth, withyear=True):
        self.year, self.month = theyear, themonth
        return super().formatmonth(theyear, themonth)

    def formatmonthname(self, theyear, themonth, withyear=True):
        dt = datetime(theyear, themonth, 1)
        return f'<tr><th colspan="7" class="month">{date_format(dt, format="b Y")}</th></tr>'

    def formatweekday(self, day):
        return f'<th class="{self.cssclasses[day]}">{_(day_abbr[day])}</th>'

    def formatday(self, day, weekday):
        if day != 0:
            cssclass = self.cssclasses[weekday]
            if date.today() == date(self.year, self.month, day):
                cssclass += " calendar-today"
            if day in self.shifts:
                cssclass += " filled"
                body = ["<br />"]
                for shift in self.shifts[day]:
                    body.append(f'<a href="{shift.event.get_absolute_url()}">')
                    body.append(shift.event.title)
                    body.append("</a><br />")
                return self.day_cell(cssclass, f"{day} {''.join(body)}")
            return self.day_cell(cssclass, day)
        return self.day_cell("noday", "&nbsp;")

    def day_cell(self, cssclass, body):
        return f'<td class="calendar-row-height p-1 break-word {cssclass}">{body}</td>'



import uuid
from collections import Counter
from functools import cached_property
from itertools import groupby
from operator import itemgetter

from django import forms
from django.core.exceptions import ValidationError
from django.template.loader import get_template
from django.utils.translation import gettext_lazy as _
from django_select2.forms import Select2MultipleWidget
from dynamic_preferences.registries import global_preferences_registry

from ephios.core.models import AbstractParticipation, Qualification
from ephios.core.signup.disposition import BaseDispositionParticipationForm
from ephios.core.signup.methods import (
    BaseSignupForm,
    BaseSignupMethod,
    BaseSignupView,
    ParticipantUnfitError,
)
from ephios.core.signup.participants import AbstractParticipant

NO_SECTION_UUID = "other"


def sections_participant_qualifies_for(sections, participant: AbstractParticipant):
    available_qualification_ids = set(q.id for q in participant.collect_all_qualifications())
    return [
        section
        for section in sections
        if set(section["qualifications"]) <= available_qualification_ids
    ]


class SectionBasedDispositionParticipationForm(BaseDispositionParticipationForm):
    disposition_participation_template = "basesignup/section_based/fragment_participation.html"

    section = forms.ChoiceField(
        label=_("Section"),
        required=False,  # only required if participation is confirmed
        widget=forms.Select(
            attrs={"data-show-for-state": str(AbstractParticipation.States.CONFIRMED)}
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        sections = self.shift.signup_method.configuration.sections
        qualified_sections = list(
            sections_participant_qualifies_for(
                sections,
                self.instance.participant,
            )
        )
        unqualified_sections = [
            section for section in sections if section not in qualified_sections
        ]
        self.fields["section"].choices = [("", "---")]
        if qualified_sections:
            self.fields["section"].choices += [
                (
                    _("qualified"),
                    [(section["uuid"], section["title"]) for section in qualified_sections],
                )
            ]
        if unqualified_sections:
            self.fields["section"].choices += [
                (
                    _("unqualified"),
                    [(section["uuid"], section["title"]) for section in unqualified_sections],
                )
            ]
        if preferred_section_uuid := self.instance.data.get("preferred_section_uuid"):
            self.fields["section"].initial = preferred_section_uuid
            self.preferred_section = next(
                filter(lambda section: section["uuid"] == preferred_section_uuid, sections), None
            )
        if initial := self.instance.data.get("dispatched_section_uuid"):
            self.fields["section"].initial = initial

    def clean(self):
        super().clean()
        if (
            self.cleaned_data["state"] == AbstractParticipation.States.CONFIRMED
            and not self.cleaned_data["section"]
        ):
            self.add_error(
                "section",
                ValidationError(_("You must select a section when confirming a participation.")),
            )

    def save(self, commit=True):
        self.instance.data["dispatched_section_uuid"] = self.cleaned_data["section"]
        super().save(commit)


class SectionForm(forms.Form):
    title = forms.CharField(label=_("Title"), required=True)
    qualifications = forms.ModelMultipleChoiceField(
        label=_("Required Qualifications"),
        queryset=Qualification.objects.all(),
        widget=Select2MultipleWidget,
        required=False,
    )
    min_count = forms.IntegerField(label=_("min amount"), min_value=0, required=True)
    max_count = forms.IntegerField(label=_("max amount"), min_value=1, required=False)
    uuid = forms.CharField(widget=forms.HiddenInput, required=False)

    def clean_uuid(self):
        return self.cleaned_data.get("uuid") or uuid.uuid4()


SectionsFormset = forms.formset_factory(
    SectionForm, can_delete=True, min_num=1, validate_min=1, extra=0
)


class SectionBasedConfigurationForm(BaseSignupMethod.configuration_form_class):
    choose_preferred_section = forms.BooleanField(
        label=_("Participants must provide a preferred section"),
        help_text=_("This only makes sense if you configure multiple sections."),
        widget=forms.CheckboxInput,
        required=False,
        initial=False,
    )
    sections = forms.Field(
        label=_("Structure"),
        widget=forms.HiddenInput,
        required=False,
    )

    def __init__(self, data=None, **kwargs):
        super().__init__(data, **kwargs)
        self.sections_formset = SectionsFormset(
            data=data,
            initial=self.initial.get("sections", []),
            prefix="sections",
        )

    def clean_sections(self):
        if not self.sections_formset.is_valid():
            raise ValidationError(_("The sections aren't configured correctly."))

        sections = [
            {
                key: cleaned_data[key]
                for key in ("title", "qualifications", "min_count", "max_count", "uuid")
            }
            for cleaned_data in self.sections_formset.cleaned_data
            if cleaned_data and not cleaned_data.get("DELETE")
        ]
        return sections

    @staticmethod
    def format_sections(value):
        return ", ".join(section["title"] for section in value)


class SectionSignupForm(BaseSignupForm):
    preferred_section_uuid = forms.ChoiceField(
        label=_("Preferred Section"),
        widget=forms.RadioSelect,
        required=False,
        # choices are later set as (uuid, title) of section
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["preferred_section_uuid"].initial = self.instance.data.get(
            "preferred_section_uuid"
        )
        self.fields["preferred_section_uuid"].required = (
            self.data.get("signup_choice") == "sign_up"
            and self.method.configuration.choose_preferred_section
        )
        self.fields["preferred_section_uuid"].choices = [
            (section["uuid"], section["title"])
            for section in self.sections_participant_qualifies_for
        ]
        unqualified = [
            section
            for section in self.method.configuration.sections
            if section not in self.sections_participant_qualifies_for
        ]
        if unqualified:
            self.fields["preferred_section_uuid"].help_text = _(
                "You don't qualify for {qualifications}."
            ).format(qualifications=", ".join(str(section["title"]) for section in unqualified))

    def save(self, commit=True):
        self.instance.data["preferred_section_uuid"] = self.cleaned_data["preferred_section_uuid"]
        return super().save(commit)

    @cached_property
    def sections_participant_qualifies_for(self):
        return sections_participant_qualifies_for(
            self.method.configuration.sections, self.participant
        )


class SectionBasedSignupView(BaseSignupView):
    form_class = SectionSignupForm


class SectionBasedSignupMethod(BaseSignupMethod):
    slug = "section_based"
    verbose_name = _("Apply for sections")
    description = _(
        """This method lets you define sections for which people can choose from.
        Sections contain qualifications that helpers need to fulfil."""
    )
    registration_button_text = _("Request")
    signup_success_message = _("You have successfully requested a participation for {shift}.")
    signup_error_message = _("Requesting a participation failed: {error}")
    signup_view_class = SectionBasedSignupView
    disposition_participation_form_class = SectionBasedDispositionParticipationForm

    configuration_form_class = SectionBasedConfigurationForm
    shift_state_template_name = "basesignup/section_based/fragment_state.html"

    def _get_signup_stats_per_section(self, participations=None):
        from ephios.core.signup.methods import SignupStats

        if participations is None:
            participations = list(self.shift.participations.all())
        confirmed_counter = Counter()
        requested_counter = Counter()
        for p in participations:
            if p.state == AbstractParticipation.States.CONFIRMED:
                c = confirmed_counter
            elif p.state == AbstractParticipation.States.REQUESTED:
                c = requested_counter
            else:
                continue
            section_uuid = p.data.get("dispatched_section_uuid")
            if not section_uuid or section_uuid not in (
                section["uuid"] for section in self.configuration.sections
            ):
                section_uuid = NO_SECTION_UUID
            c[section_uuid] += 1

        d = {}
        for section in self.configuration.sections:
            section_uuid = section["uuid"]
            min_count = section.get("min_count")
            max_count = section.get("max_count")
            d[section_uuid] = SignupStats(
                requested_count=requested_counter[section_uuid],
                confirmed_count=confirmed_counter[section_uuid],
                missing=(max(min_count - confirmed_counter[section_uuid], 0) if min_count else 0),
                free=(max(max_count - confirmed_counter[section_uuid], 0) if max_count else None),
                min_count=min_count,
                max_count=max_count,
            )

        # Participations not assigned to a section are extra, so max and free are explicitly zero.
        # We do not offset missing places in other sections, as qualifications etc. might not match.
        # Disposition will always be required to resolve unassigned participations.
        d[NO_SECTION_UUID] = SignupStats(
            requested_count=requested_counter[NO_SECTION_UUID],
            confirmed_count=confirmed_counter[NO_SECTION_UUID],
            missing=0,
            free=0,
            min_count=None,
            max_count=0,
        )

        return d

    def get_signup_stats(self):
        from ephios.core.signup.methods import SignupStats

        participations = list(self.shift.participations.all())

        signup_stats = SignupStats.ZERO
        for stats in self._get_signup_stats_per_section(participations).values():
            signup_stats += stats

        return signup_stats

    @staticmethod
    def check_qualification(method, participant):
        if not sections_participant_qualifies_for(method.configuration.sections, participant):
            return ParticipantUnfitError(_("You are not qualified."))

    @property
    def _signup_checkers(self):
        return super()._signup_checkers + [self.check_qualification]

    # pylint: disable=arguments-differ
    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        participation.state = AbstractParticipation.States.REQUESTED
        return participation

    def render_configuration_form(self, *args, form=None, **kwargs):
        """We overwrite the template to render the formset."""
        form = form or self.get_configuration_form(*args, **kwargs)
        template = get_template("basesignup/section_based/configuration_form.html").render(
            {"form": form}
        )
        return template

    def get_shift_state_context_data(self, request, **kwargs):
        context_data = super().get_shift_state_context_data(request)
        participations = context_data["participations"]
        section_stats = self._get_signup_stats_per_section(participations)
        sections = {
            section["uuid"]: {
                "title": section["title"],
                "placeholder": section.get("min_count") or 0,
                "qualifications_label": ", ".join(
                    q.abbreviation
                    for q in Qualification.objects.filter(id__in=section["qualifications"])
                ),
                "participations": [],
                "stats": section_stats[section["uuid"]],
            }
            for section in self.configuration.sections
        }

        unsorted_participations = []
        for participation in participations:
            dispatched_uuid = participation.data.get(
                "dispatched_section_uuid"
            ) or participation.data.get("preferred_section_uuid")
            if dispatched_uuid not in sections:
                unsorted_participations.append(participation)
            else:
                sections[dispatched_uuid]["participations"].append(participation)
                sections[dispatched_uuid]["placeholder"] -= 1

        for section in sections.values():
            section["placeholder"] = list(range(max(0, section["placeholder"])))
        if unsorted_participations:
            sections[NO_SECTION_UUID] = {
                "title": _("unassigned"),
                "participations": unsorted_participations,
                "placeholder": [],
            }

        context_data["sections"] = sections
        return context_data

    def _get_sections_with_users(self):
        relevant_qualification_categories = global_preferences_registry.manager()[
            "general__relevant_qualification_categories"
        ]
        section_by_uuid = {section["uuid"]: section for section in self.configuration.sections}
        # get name and preferred section uuid for confirmed participants
        # if they have a section assigned and we have that section on record
        confirmed_participations = [
            {
                "name": str(participation.participant),
                "relevant_qualifications": ", ".join(
                    participation.participant.qualifications.filter(
                        category__in=relevant_qualification_categories
                    )
                    .order_by("category", "abbreviation")
                    .values_list("abbreviation", flat=True)
                ),
                "uuid": dispatched_section_uuid,
            }
            for participation in self.shift.participations.filter(
                state=AbstractParticipation.States.CONFIRMED
            )
            if (dispatched_section_uuid := participation.data.get("dispatched_section_uuid"))
            and dispatched_section_uuid in section_by_uuid
        ]
        # group by section and do some stats
        sections_with_users = [
            (
                section_by_uuid.pop(uuid),
                [[user["name"], user["relevant_qualifications"]] for user in group],
            )
            for uuid, group in groupby(
                sorted(confirmed_participations, key=itemgetter("uuid")), itemgetter("uuid")
            )
        ]
        # add sections without participants
        sections_with_users += [(section, None) for section in section_by_uuid.values()]
        return sections_with_users

    def get_participation_display(self):
        confirmed_sections_with_users = self._get_sections_with_users()
        participation_display = []
        for section, users in confirmed_sections_with_users:
            if users:
                participation_display += [[user[0], user[1], section["title"]] for user in users]
            if not users or len(users) < section["min_count"]:
                required_qualifications = ", ".join(
                    Qualification.objects.filter(pk__in=section["qualifications"]).values_list(
                        "abbreviation", flat=True
                    )
                )
                participation_display += [["", required_qualifications, section["title"]]] * (
                    section["min_count"] - (len(users) if users else 0)
                )
        return participation_display



# Generated by Django 3.1.7 on 2021-03-23 22:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0006_notification"),
    ]

    operations = [
        migrations.AddField(
            model_name="eventtype",
            name="color",
            field=models.CharField(default="#343a40", max_length=7, verbose_name="color"),
        ),
    ]



# Generated by Django 3.1.5 on 2021-03-05 17:55

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("core", "0003_migrate_content_type_to_core_20210219_1642"),
    ]

    operations = [
        migrations.AlterField(
            model_name="abstractparticipation",
            name="polymorphic_ctype",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="polymorphic_core.abstractparticipation_set+",
                to="contenttypes.contenttype",
            ),
        ),
        migrations.AlterField(
            model_name="abstractparticipation",
            name="shift",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="participations",
                to="core.shift",
                verbose_name="shift",
            ),
        ),
    ]



import logging
import operator
from collections import Counter

from django.db.models import Q

from ephios.core.consequences import QualificationConsequenceHandler
from ephios.core.models import AbstractParticipation, LocalParticipation
from ephios.plugins.eventautoqualification.models import EventAutoQualificationConfiguration

logger = logging.getLogger(__name__)


def create_qualification_consequence(sender, participation, **kwargs):
    event = participation.shift.event
    try:
        mode = event.auto_qualification_config.mode
    except EventAutoQualificationConfiguration.DoesNotExist:
        return

    # check if the required shift participations exist
    requirements_met = False
    if mode == EventAutoQualificationConfiguration.Modes.ANY_SHIFT:
        requirements_met = True
    elif mode == EventAutoQualificationConfiguration.Modes.LAST_SHIFT:
        requirements_met = (
            participation.shift
            == sorted(event.shifts.all(), key=operator.attrgetter("end_time"))[-1]
        )
    elif mode == EventAutoQualificationConfiguration.Modes.EVERY_SHIFT:
        # count participant hashes in finished participations (and this one)
        participant_hash_counter = Counter(
            hash(participation.participant)
            for participation in AbstractParticipation.objects.filter(
                Q(finished=True) | Q(id=participation.id)
            )
        )
        # if the amount of this participant in confirmed past shifts is equal to the the count of all shifts
        # then this participant attended every shift
        requirements_met = participant_hash_counter[hash(participation.participant)] == len(
            event.shifts.all()
        )

    if requirements_met:
        if not isinstance(participation, LocalParticipation):
            logger.warning(
                "Cannot create an automatic qualification consequence for non-local participants."
            )
            return

        user = participation.user

        # skip if extent/refresh only but the user does not have a grant
        if (
            event.auto_qualification_config.extend_only
            and not event.auto_qualification_config.qualification_id
            in user.qualification_grants.values_list("qualification_id", flat=True)
        ):
            return

        consequence = QualificationConsequenceHandler.create(
            user=user,
            qualification=event.auto_qualification_config.qualification,
            expires=event.auto_qualification_config.expiration_date,
            shift=participation.shift,
        )

        if not event.auto_qualification_config.needs_confirmation:
            consequence.confirm(user=None)



import typing

from django import forms
from django.utils.translation import gettext_lazy as _
from django_select2.forms import Select2MultipleWidget
from dynamic_preferences.registries import global_preferences_registry

from ephios.core.models import AbstractParticipation, Qualification
from ephios.core.signup.methods import (
    BaseSignupMethod,
    ParticipantUnfitError,
    SignupDisallowedError,
)

_Base = BaseSignupMethod if typing.TYPE_CHECKING else object


class MinMaxParticipantsMixin(_Base):
    @property
    def _signup_checkers(self):
        return super()._signup_checkers + [self.check_maximum_number_of_participants]

    @staticmethod
    def check_maximum_number_of_participants(method, participant):
        if (
            not method.uses_requested_state
            and method.configuration.maximum_number_of_participants is not None
        ):
            current_count = len(
                [
                    participation
                    for participation in method.shift.participations.all()
                    if participation.state == AbstractParticipation.States.CONFIRMED
                ]
            )
            if current_count >= method.configuration.maximum_number_of_participants:
                return SignupDisallowedError(_("The maximum number of participants is reached."))

    def get_participant_count_bounds(self):
        return (
            self.configuration.minimum_number_of_participants,
            self.configuration.maximum_number_of_participants,
        )

    @property
    def configuration_form_class(self):
        class ConfigurationForm(super().configuration_form_class):
            minimum_number_of_participants = forms.IntegerField(
                label=_("Minimum number of participants"), min_value=0, required=False
            )
            maximum_number_of_participants = forms.IntegerField(
                label=_("Maximum number of participants"), min_value=1, required=False
            )

        return ConfigurationForm

    def get_shift_state_context_data(self, request, **kwargs):
        context_data = super().get_shift_state_context_data(request, **kwargs)
        context_data["empty_spots"] = range(
            max(
                0,
                (getattr(self.configuration, "minimum_number_of_participants", None) or 0)
                - len(context_data["participations"]),
            )
        )
        return context_data


class QualificationsRequiredSignupMixin(_Base):
    def __init__(self, shift, **kwargs):
        super().__init__(shift, **kwargs)
        if shift is not None:
            self.configuration.required_qualifications = Qualification.objects.filter(
                pk__in=self.configuration.required_qualification_ids
            )

    @property
    def _participant_checkers(self):
        return super()._participant_checkers + [self.check_qualification]

    @staticmethod
    def check_qualification(method, participant):
        if not participant.has_qualifications(method.configuration.required_qualifications):
            return ParticipantUnfitError(_("You are not qualified."))

    @property
    def configuration_form_class(self):
        class ConfigurationForm(super().configuration_form_class):
            required_qualification_ids = forms.ModelMultipleChoiceField(
                label=_("Required Qualifications"),
                queryset=Qualification.objects.all(),
                widget=Select2MultipleWidget,
                required=False,
                initial=[],
            )

            @staticmethod
            def format_required_qualification_ids(ids):
                return ", ".join(
                    Qualification.objects.filter(id__in=ids).values_list("title", flat=True)
                )

        return ConfigurationForm

    def get_shift_state_context_data(self, request, **kwargs):
        return super().get_shift_state_context_data(
            request,
            **kwargs,
            required_qualification=", ".join(
                q.abbreviation for q in self.configuration.required_qualifications
            ),
        )


class RenderParticipationPillsShiftStateMixin:
    shift_state_template_name = "basesignup/fragment_participation_pills_shift_state.html"


class QualificationMinMaxBaseSignupMethod(
    RenderParticipationPillsShiftStateMixin,
    QualificationsRequiredSignupMixin,
    MinMaxParticipantsMixin,
    BaseSignupMethod,
):
    @property
    def slug(self):
        raise NotImplementedError()

    def get_participation_display(self):
        relevant_qualification_categories = global_preferences_registry.manager()[
            "general__relevant_qualification_categories"
        ]
        participation_info = [
            [
                f"{participant.first_name} {participant.last_name}",
                ", ".join(
                    participant.qualifications.filter(
                        category__in=relevant_qualification_categories
                    )
                    .order_by("category", "title")
                    .values_list("title", flat=True)
                ),
            ]
            for participant in self.shift.get_participants()
        ]
        min_count, max_count = self.get_participant_count_bounds()
        rendered_count = max(min_count or 0, max_count or 0)
        if len(participation_info) < rendered_count:
            required_qualifications = self.configuration.required_qualifications
            qualifications_display = (
                ", ".join(required_qualifications.values_list("title", flat=True))
                if required_qualifications
                else "-"
            )
            participation_info += [["", qualifications_display]] * (
                rendered_count - len(participation_info)
            )
        return participation_info



from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext_lazy

from ephios.core.models import AbstractParticipation
from ephios.plugins.basesignup.signup.common import QualificationMinMaxBaseSignupMethod


class InstantConfirmationSignupMethod(QualificationMinMaxBaseSignupMethod):
    slug = "instant_confirmation"
    verbose_name = _("Instant Confirmation")
    description = _("""This method instantly confirms every signup after it was requested.""")
    uses_requested_state = False
    registration_button_text = pgettext_lazy("signup button text", "Participate")

    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        participation.state = AbstractParticipation.States.CONFIRMED
        return participation



from django import template

register = template.Library()


@register.simple_tag(takes_context=True)
def param_replace(context, **kwargs):
    params = context["request"].GET.copy()
    for key, value in kwargs.items():
        if not value and key in params:
            del params[key]
        else:
            params[key] = str(value)
    return params.urlencode(safe="[]")



from django.contrib import messages
from django.contrib.auth.forms import PasswordResetForm
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.auth.models import Group
from django.contrib.messages.views import SuccessMessageMixin
from django.db.models import Prefetch
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    FormView,
    ListView,
    TemplateView,
    UpdateView,
)
from django.views.generic.detail import SingleObjectMixin
from dynamic_preferences.registries import global_preferences_registry

from ephios.core.forms.users import (
    GroupForm,
    QualificationGrantFormset,
    UserNotificationPreferenceForm,
    UserProfileForm,
)
from ephios.core.models import QualificationGrant, UserProfile
from ephios.core.services.notifications.types import (
    NewProfileNotification,
    ProfileUpdateNotification,
)
from ephios.extra.mixins import CustomPermissionRequiredMixin


class ProfileView(LoginRequiredMixin, DetailView):
    def get_object(self, queryset=None):
        return self.request.user

    def get_context_data(self, **kwargs):
        return super().get_context_data(own_profile=True, **kwargs)


class UserProfileDetailView(CustomPermissionRequiredMixin, DetailView):
    model = UserProfile
    permission_required = "core.view_userprofile"


class UserProfileListView(CustomPermissionRequiredMixin, ListView):
    model = UserProfile
    permission_required = "core.view_userprofile"
    ordering = "last_name"

    def get_queryset(self):
        global_preferences = global_preferences_registry.manager()
        categories = global_preferences["general__relevant_qualification_categories"]
        qs = UserProfile.objects.all()
        for category in categories:
            qs = qs.prefetch_related(
                Prefetch(
                    "qualification_grants",
                    queryset=QualificationGrant.objects.filter(
                        qualification__category=category,
                    ).select_related("qualification"),
                    to_attr=f"qualifications_for_category_{category.pk}",
                )
            )
        return qs.order_by("last_name", "first_name")


class UserProfileCreateView(CustomPermissionRequiredMixin, TemplateView):
    template_name = "core/userprofile_form.html"
    permission_required = "core.add_userprofile"
    model = UserProfile

    def get_context_data(self, **kwargs):
        kwargs.setdefault(
            "userprofile_form", UserProfileForm(self.request.POST or None, request=self.request)
        )
        kwargs.setdefault(
            "qualification_formset", QualificationGrantFormset(self.request.POST or None)
        )
        return super().get_context_data(**kwargs)

    def post(self, request, *args, **kwargs):
        userprofile_form = UserProfileForm(self.request.POST, request=request)
        qualification_formset = QualificationGrantFormset(self.request.POST)
        if all((userprofile_form.is_valid(), qualification_formset.is_valid())):
            userprofile = userprofile_form.save()
            qualification_formset.instance = userprofile
            qualification_formset.save()
            messages.success(
                self.request,
                _("User {name} ({email}) added successfully.").format(
                    name=userprofile.get_full_name(), email=userprofile.email
                ),
            )
            if userprofile.is_active:
                NewProfileNotification.send(userprofile)
            return redirect(reverse("core:userprofile_list"))
        return self.render_to_response(
            self.get_context_data(
                userprofile_form=userprofile_form, qualification_formset=qualification_formset
            )
        )


class UserProfileUpdateView(CustomPermissionRequiredMixin, SingleObjectMixin, TemplateView):
    model = UserProfile
    permission_required = "core.change_userprofile"
    template_name = "core/userprofile_form.html"

    def get_userprofile_form(self):
        return UserProfileForm(
            self.request.POST or None,
            initial={
                "groups": self.get_object().groups.all(),
            },
            instance=self.object,
            request=self.request,
        )

    def get_qualification_formset(self):
        return QualificationGrantFormset(self.request.POST or None, instance=self.object)

    def get_context_data(self, **kwargs):
        self.object = self.get_object()
        kwargs.setdefault("userprofile_form", self.get_userprofile_form())
        kwargs.setdefault("qualification_formset", self.get_qualification_formset())
        return super().get_context_data(**kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        userprofile_form = self.get_userprofile_form()
        qualification_formset = self.get_qualification_formset()
        if all((userprofile_form.is_valid(), qualification_formset.is_valid())):
            userprofile = userprofile_form.save()
            qualification_formset.save()
            messages.success(
                self.request,
                _("User {name} ({user}) updated successfully.").format(
                    name=self.object.get_full_name(), user=self.object
                ),
            )
            if request.user != userprofile:
                ProfileUpdateNotification.send(userprofile)
            return redirect(reverse("core:userprofile_list"))

        return self.render_to_response(
            self.get_context_data(
                userprofile_form=userprofile_form, qualification_formset=qualification_formset
            )
        )


class UserProfileDeleteView(CustomPermissionRequiredMixin, DeleteView):
    model = UserProfile
    permission_required = "core.delete_userprofile"
    template_name = "core/userprofile_confirm_delete.html"

    def get_success_url(self):
        messages.info(
            self.request,
            _("The user {name} ({user}) was deleted.").format(
                name=self.object.get_full_name(), user=self.object
            ),
        )
        return reverse("core:userprofile_list")


class UserProfilePasswordResetView(CustomPermissionRequiredMixin, SingleObjectMixin, TemplateView):
    model = UserProfile
    permission_required = "core.change_userprofile"
    template_name = "core/userprofile_confirm_password_reset.html"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.object = self.get_object()

    def post(self, request, *args, **kwargs):
        if request.POST.get("confirm"):
            form = PasswordResetForm(
                {
                    "email": self.object.email,
                }
            )
            if form.is_valid():
                form.save(request=request)
                messages.info(
                    request,
                    _("The user's password has been reset. An email was sent to {email}.").format(
                        email=self.object.email
                    ),
                )
            else:
                messages.error(
                    request,
                    _("No valid email address ({email}). The password has not been reset.").format(
                        email=self.object.email
                    ),
                )
            return redirect(reverse("core:userprofile_list"))
        return self.render_to_response({"userprofile": self.object})


class UserProfileNotificationsView(LoginRequiredMixin, SuccessMessageMixin, FormView):
    template_name = "core/userprofile_notifications.html"
    success_message = _("Settings succesfully saved.")

    def get_form(self, form_class=None):
        return UserNotificationPreferenceForm(self.request.POST or None, user=self.request.user)

    def get_success_url(self):
        return reverse("core:profile_notifications")

    def form_valid(self, form):
        form.update_preferences()
        return super().form_valid(form)


class GroupListView(CustomPermissionRequiredMixin, ListView):
    model = Group
    permission_required = "auth.view_group"
    template_name = "core/group_list.html"
    ordering = "name"


class GroupCreateView(CustomPermissionRequiredMixin, CreateView):
    model = Group
    permission_required = "auth.add_group"
    accept_object_perms = False
    template_name = "core/group_form.html"
    form_class = GroupForm

    def get_success_url(self):
        messages.success(
            self.request, _('Group "{group}" created successfully.').format(group=self.object)
        )
        return reverse("core:group_list")


class GroupUpdateView(CustomPermissionRequiredMixin, UpdateView):
    model = Group
    permission_required = "auth.change_group"
    template_name = "core/group_form.html"
    form_class = GroupForm

    def get_success_url(self):
        messages.success(
            self.request, _('Group "{group}" updated successfully.').format(group=self.object)
        )
        return reverse("core:group_list")


class GroupDeleteView(CustomPermissionRequiredMixin, DeleteView):
    model = Group
    permission_required = "auth.delete_group"
    template_name = "core/group_confirm_delete.html"

    def get_success_url(self):
        messages.info(self.request, _('The group "{group}" was deleted.').format(group=self.object))
        return reverse("core:group_list")



from django.urls import include, path

from ephios.plugins.guests.views import (
    GuestEventDetailView,
    GuestRegistrationView,
    GuestUserShiftActionView,
)

app_name = "guests"
urlpatterns = [
    path(
        "guests/register/<int:event_id>/<slug:public_signup_token>/",
        GuestRegistrationView.as_view(),
        name="register",
    ),
    path(
        "guests/<guest_access_token>/",
        include(
            [
                path("", GuestEventDetailView.as_view(), name="event_detail"),
                path(
                    "shifts/<int:pk>/signup-action/",
                    GuestUserShiftActionView.as_view(),
                    name="signup_action",
                ),
            ]
        ),
    ),
]



# Generated by Django 3.1.7 on 2021-03-27 14:36

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import ephios.modellogging.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="LogEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("content_object_id", models.PositiveIntegerField(db_index=True)),
                ("attached_to_object_id", models.PositiveIntegerField(db_index=True)),
                ("datetime", models.DateTimeField(auto_now_add=True)),
                (
                    "action_type",
                    models.CharField(
                        choices=[
                            (
                                ephios.modellogging.models.InstanceActionType["CREATE"],
                                ephios.modellogging.models.InstanceActionType["CREATE"],
                            ),
                            (
                                ephios.modellogging.models.InstanceActionType["CHANGE"],
                                ephios.modellogging.models.InstanceActionType["CHANGE"],
                            ),
                            (
                                ephios.modellogging.models.InstanceActionType["DELETE"],
                                ephios.modellogging.models.InstanceActionType["DELETE"],
                            ),
                        ],
                        max_length=255,
                    ),
                ),
                ("request_id", models.CharField(blank=True, max_length=36, null=True)),
                (
                    "data",
                    models.JSONField(
                        default=dict, encoder=ephios.modellogging.models.LogJSONEncoder
                    ),
                ),
                (
                    "attached_to_object_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="logs_for_me",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="logs_about_me",
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="logging_entries",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "ordering": ("-datetime", "-id"),
            },
        ),
    ]



import uuid

from .log import log_request_store


class LoggingRequestMiddleware:
    """
    This middleware sets request as a local thread variable, making it
    available to the logging utilities to allow tracking of the
    authenticated user making a change.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        log_request_store.request = request
        log_request_store.request_id = str(uuid.uuid4())

        response = self.get_response(request)

        del log_request_store.request
        del log_request_store.request_id
        return response



from .events import (
    AbstractParticipation,
    Event,
    EventType,
    EventTypePreference,
    LocalParticipation,
    Shift,
)
from .users import (
    Consequence,
    Notification,
    Qualification,
    QualificationCategory,
    QualificationGrant,
    UserProfile,
    WorkingHours,
)



from django.urls import path

from ephios.core import pdf
from ephios.core.ical import user_event_feed_view
from ephios.core.signup.disposition import (
    AddPlaceholderParticipantView,
    AddUserView,
    DispositionView,
)
from ephios.core.views.accounts import (
    GroupCreateView,
    GroupDeleteView,
    GroupListView,
    GroupUpdateView,
    ProfileView,
    UserProfileCreateView,
    UserProfileDeleteView,
    UserProfileDetailView,
    UserProfileListView,
    UserProfileNotificationsView,
    UserProfilePasswordResetView,
    UserProfileUpdateView,
)
from ephios.core.views.bulk import EventBulkDeleteView
from ephios.core.views.consequences import ConsequenceUpdateView, WorkingHourRequestView
from ephios.core.views.event import (
    EventActivateView,
    EventArchiveView,
    EventCopyView,
    EventCreateView,
    EventDeleteView,
    EventDetailView,
    EventListTypeSettingView,
    EventNotificationView,
    EventUpdateView,
    HomeView,
    RRuleOccurrenceView,
    current_event_list_view,
)
from ephios.core.views.eventtype import (
    EventTypeCreateView,
    EventTypeDeleteView,
    EventTypeListView,
    EventTypeUpdateView,
)
from ephios.core.views.log import LogView
from ephios.core.views.pwa import manifest, offline, serviceworker
from ephios.core.views.settings import GeneralSettingsView
from ephios.core.views.shift import (
    ShiftConfigurationFormView,
    ShiftCreateView,
    ShiftDeleteView,
    ShiftUpdateView,
)
from ephios.core.views.signup import LocalUserShiftActionView

app_name = "core"
urlpatterns = [
    path("", HomeView.as_view(), name="index"),
    path("events/", current_event_list_view, name="event_list"),
    path(
        "events/<int:pk>/edit/",
        EventUpdateView.as_view(),
        name="event_edit",
    ),
    path(
        "events/<int:pk>/delete/",
        EventDeleteView.as_view(),
        name="event_delete",
    ),
    path(
        "events/<int:pk>-<slug:slug>/",
        EventDetailView.as_view(),
        name="event_detail",
    ),
    path(
        "events/<int:pk>/createshift/",
        ShiftCreateView.as_view(),
        name="event_createshift",
    ),
    path(
        "events/<int:pk>/activate/",
        EventActivateView.as_view(),
        name="event_activate",
    ),
    path(
        "events/<int:pk>/notifications/",
        EventNotificationView.as_view(),
        name="event_notifications",
    ),
    path("events/<int:pk>/pdf/", pdf.EventDetailPDFView.as_view(), name="event_detail_pdf"),
    path(
        "events/<int:pk>/copy",
        EventCopyView.as_view(),
        name="event_copy",
    ),
    path(
        "events/create/<int:type>/",
        EventCreateView.as_view(),
        name="event_create",
    ),
    path(
        "events/past/",
        EventArchiveView.as_view(),
        name="event_list_past",
    ),
    path(
        "events/delete",
        EventBulkDeleteView.as_view(),
        name="event_bulk_delete",
    ),
    path(
        "shifts/<int:pk>/signup/",
        LocalUserShiftActionView.as_view(),
        name="signup_action",
    ),
    path(
        "shifts/<int:pk>/edit/",
        ShiftUpdateView.as_view(),
        name="shift_edit",
    ),
    path(
        "shifts/<int:pk>/delete/",
        ShiftDeleteView.as_view(),
        name="shift_delete",
    ),
    path(
        "signup_method_configuration_form/<int:event_id>/<slug:slug>/",
        ShiftConfigurationFormView.as_view(),
        name="signupmethod_configurationform",
    ),
    path(
        "shifts/<int:pk>/disposition/",
        DispositionView.as_view(),
        name="shift_disposition",
    ),
    path(
        "shifts/<int:pk>/disposition/add-user/",
        AddUserView.as_view(),
        name="shift_disposition_add_user",
    ),
    path(
        "shifts/<int:pk>/disposition/add-placeholder/",
        AddPlaceholderParticipantView.as_view(),
        name="shift_disposition_add_placeholder",
    ),
    path("calendar/<str:calendar_token>/", user_event_feed_view, name="user_event_feed"),
    path(
        "extra/rruleoccurrence",
        RRuleOccurrenceView.as_view(),
        name="rrule_occurrences",
    ),
    path("settings/general/", GeneralSettingsView.as_view(), name="settings_general"),
    path("settings/eventtypes/", EventTypeListView.as_view(), name="settings_eventtype_list"),
    path(
        "settings/eventtypes/create/",
        EventTypeCreateView.as_view(),
        name="settings_eventtype_create",
    ),
    path(
        "settings/eventtypes/<int:pk>/edit/",
        EventTypeUpdateView.as_view(),
        name="settings_eventtype_edit",
    ),
    path(
        "settings/eventtypes/<int:pk>/delete/",
        EventTypeDeleteView.as_view(),
        name="settings_eventtype_delete",
    ),
    path("profile/", ProfileView.as_view(), name="profile"),
    path(
        "profile/notifications",
        UserProfileNotificationsView.as_view(),
        name="profile_notifications",
    ),
    path("groups/", GroupListView.as_view(), name="group_list"),
    path("groups/<int:pk>/edit", GroupUpdateView.as_view(), name="group_edit"),
    path("groups/<int:pk>/delete", GroupDeleteView.as_view(), name="group_delete"),
    path("groups/create", GroupCreateView.as_view(), name="group_add"),
    path(
        "users/",
        UserProfileListView.as_view(),
        name="userprofile_list",
    ),
    path(
        "users/<int:pk>/edit",
        UserProfileUpdateView.as_view(),
        name="userprofile_edit",
    ),
    path(
        "users/<int:pk>/delete",
        UserProfileDeleteView.as_view(),
        name="userprofile_delete",
    ),
    path(
        "users/<int:pk>/password_reset",
        UserProfilePasswordResetView.as_view(),
        name="userprofile_password_reset",
    ),
    path(
        "users/create/",
        UserProfileCreateView.as_view(),
        name="userprofile_create",
    ),
    path(
        "users/<int:pk>",
        UserProfileDetailView.as_view(),
        name="userprofile_detail",
    ),
    path(
        "consequences/<int:pk>/edit",
        ConsequenceUpdateView.as_view(),
        name="consequence_edit",
    ),
    path(
        "profile/requestworkinghour",
        WorkingHourRequestView.as_view(),
        name="request_workinghour",
    ),
    path("log/", LogView.as_view(), name="log"),
    path("manifest.json", manifest, name="pwa_manifest"),
    path("serviceworker.js", serviceworker, name="pwa_serviceworker"),
    path("offline/", offline, name="pwa_offline"),
    path(
        "events/list_type_setting",
        EventListTypeSettingView.as_view(),
        name="event_list_type_setting",
    ),
]



import logging
import traceback

from django.conf import settings
from django.core.mail import EmailMultiAlternatives, mail_admins
from django.utils.translation import gettext_lazy as _
from webpush import send_user_notification

from ephios.core.models.users import Notification

logger = logging.getLogger(__name__)


def installed_notification_backends():
    from ephios.core.signals import register_notification_backends

    for _, backends in register_notification_backends.send_to_all_plugins(None):
        yield from (b() for b in backends)


def enabled_notification_backends():
    from ephios.core.signals import register_notification_backends

    for _, backends in register_notification_backends.send(None):
        yield from (b() for b in backends)


def send_all_notifications():
    for backend in installed_notification_backends():
        for notification in Notification.objects.filter(failed=False):
            if backend.can_send(notification) and backend.user_prefers_sending(notification):
                try:
                    backend.send(notification)
                except Exception as e:  # pylint: disable=broad-except
                    if settings.DEBUG:
                        raise e
                    notification.failed = True
                    notification.save()
                    mail_admins(
                        "Notification sending failed",
                        f"Notification: {notification}\nException: {e}\n{traceback.format_exc()}",
                    )
                    logger.warning(
                        f"Notification sending failed for notification object #{notification.pk} ({notification}) for backend {backend} with {e}"
                    )
    Notification.objects.filter(failed=False).delete()


class AbstractNotificationBackend:
    @property
    def slug(self):
        return NotImplementedError

    @property
    def title(self):
        return NotImplementedError

    @classmethod
    def can_send(cls, notification):
        return notification.user is not None

    @classmethod
    def user_prefers_sending(cls, notification):
        if notification.notification_type.unsubscribe_allowed and notification.user is not None:
            if not notification.user.is_active:
                return False
            backends = notification.user.preferences["notifications__notifications"].get(
                notification.slug
            )
            if backends is not None:
                return cls.slug in backends
        return True

    @classmethod
    def send(cls, notification: Notification):
        raise NotImplementedError


class EmailNotificationBackend(AbstractNotificationBackend):
    slug = "ephios_backend_email"
    title = _("via email")

    @classmethod
    def can_send(cls, notification):
        return notification.user is not None or "email" in notification.data

    @classmethod
    def _get_mailaddress(cls, notification):
        return notification.user.email if notification.user else notification.data.get("email")

    @classmethod
    def send(cls, notification):
        email = EmailMultiAlternatives(
            to=[cls._get_mailaddress(notification)],
            subject=notification.subject,
            body=notification.as_plaintext(),
        )
        email.attach_alternative(notification.as_html(), "text/html")
        email.send()


class WebPushNotificationBackend(AbstractNotificationBackend):
    slug = "ephios_backend_webpush"
    title = _("via push notification")

    @classmethod
    def send(cls, notification):
        payload = {
            "head": str(notification.subject),
            "body": notification.as_plaintext(),
        }
        if url := notification.get_url():
            payload["url"] = url
        send_user_notification(user=notification.user, payload=payload, ttl=1000)


CORE_NOTIFICATION_BACKENDS = [EmailNotificationBackend, WebPushNotificationBackend]



from django import forms
from django.db.models import Q
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ephios.core.models import AbstractParticipation, Shift
from ephios.core.signup.methods import ActionDisallowedError, BaseSignupMethod
from ephios.plugins.basesignup.signup.common import RenderParticipationPillsShiftStateMixin


class CoupledSignupMethod(RenderParticipationPillsShiftStateMixin, BaseSignupMethod):
    slug = "coupled"
    verbose_name = _("coupled to another shift")
    description = _("""This method mirrors signup from another shift.""")
    uses_requested_state = True

    @property
    def configuration_form_class(self):
        class ConfigurationForm(super().configuration_form_class):
            leader_shift_id = forms.ModelChoiceField(
                label=_("shift to mirror participation from"),
                required=True,
                queryset=self.event.shifts.exclude(
                    Q(pk__in=[self.shift.id] if self.shift else [])
                    | Q(signup_method_slug=self.slug)  # no chaining/no cycles!
                ),
            )

            @staticmethod
            def format_leader_shift_id(value):
                return str(Shift.objects.get(id=value) if isinstance(value, int) else value)

        return ConfigurationForm

    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        raise TypeError(f"{self.__class__} does not support signup")

    @cached_property
    def leader_shift(self):
        if isinstance(self.configuration.leader_shift_id, Shift):
            return self.configuration.leader_shift_id
        try:
            return Shift.objects.get(id=self.configuration.leader_shift_id)
        except Shift.DoesNotExist:
            return None

    @staticmethod
    def signup_is_disabled(method, participant):
        if method.leader_shift:
            return ActionDisallowedError(
                _("Participation is coupled to {}.").format(method.leader_shift)
            )
        # This is red as it requires responsibles to update the shift configuration.
        return ActionDisallowedError(
            mark_safe(
                f'<span class="text-danger">{_("Participation is coupled to another shift, but the leading shift is missing.")}</span>'
            )
        )

    @property
    def _signup_checkers(self):
        return [
            self.signup_is_disabled,
        ]

    @property
    def _decline_checkers(self):
        return [
            self.signup_is_disabled,
        ]


@receiver(pre_save)
def update_coupled_shifts(sender, instance, **kwargs):
    # make sure sender is a true subclass of AbstractParticipation
    if not issubclass(sender, AbstractParticipation) and sender != AbstractParticipation:
        return

    if instance.state == AbstractParticipation.States.GETTING_DISPATCHED:
        return

    coupled_shifts = [
        shift
        for shift in instance.shift.event.shifts.all()
        if shift.signup_method.slug == CoupledSignupMethod.slug
        and shift.signup_method.leader_shift == instance.shift
    ]

    for coupled_shift in coupled_shifts:
        # make sure a type `instance` exists also for `coupled_shift`
        coupled_participation = coupled_shift.signup_method.get_participation_for(
            instance.participant
        )
        coupled_participation.state = instance.state
        coupled_participation.save()


@receiver(post_save)
def fill_new_shift(sender, instance, **kwargs):
    # make sure sender is a true subclass of AbstractParticipation
    if not issubclass(sender, Shift) or instance.signup_method_slug != CoupledSignupMethod.slug:
        return

    if not (leader_shift := instance.signup_method.leader_shift):
        return

    for leader_participation in leader_shift.participations.exclude(
        state=AbstractParticipation.States.GETTING_DISPATCHED
    ):
        # make sure a type `instance` exists also for `coupled_shift`
        coupled_participation = instance.signup_method.get_participation_for(
            leader_participation.participant
        )
        coupled_participation.state = leader_participation.state
        coupled_participation.save()



from django.contrib.auth.mixins import AccessMixin, PermissionRequiredMixin
from django.shortcuts import redirect
from django.urls import resolve


class CustomPermissionRequiredMixin(PermissionRequiredMixin):
    """
    As of 2020-09-26, guardians permission mixin
    doesn't support the mode of operation we want, but Django's does:
    * Logged in users without permission get 403
    * not logged in users get redirected to login
    Therefore we patch Django's mixin to support object permissions
    """

    accept_global_perms = True
    accept_object_perms = True

    def get_permission_object(self):
        if hasattr(self, "permission_object"):
            return self.permission_object
        if hasattr(self, "get_object") and (obj := self.get_object()) is not None:
            return obj
        return getattr(self, "object", None)

    def has_permission(self):
        user = self.request.user
        perms = self.get_permission_required()
        if self.accept_global_perms and all(user.has_perm(perm) for perm in perms):
            return True
        if not self.accept_object_perms or (obj := self.get_permission_object()) is None:
            return False
        return all(user.has_perm(perm, obj) for perm in perms)


class StaffRequiredMixin(AccessMixin):
    """Verify that the current user is staff."""

    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_staff:
            return self.handle_no_permission()
        return super().dispatch(request, *args, **kwargs)


class CanonicalSlugDetailMixin:
    """
    A mixin that enforces a canonical slug in the url.
    If a urlpattern takes a object's pk and slug as arguments and the slug url
    argument does not equal the object's canonical slug, this mixin will
    redirect to the url containing the canonical slug.

    Taken from django-braces, licenced under BSD Licence 2.0:
    Copyright (c) Brack3t and individual contributors.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification,
    are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
           this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
           notice, this list of conditions and the following disclaimer in the
           documentation and/or other materials provided with the distribution.

        3. Neither the name of Brack3t nor the names of its contributors may be used
           to endorse or promote products derived from this software without
           specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
    ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

    def dispatch(self, request, *args, **kwargs):
        # Set up since we need to super() later instead of earlier.
        self.request = request
        self.args = args
        self.kwargs = kwargs

        # Get the current object, url slug, and
        # urlpattern name (namespace aware).
        obj = self.get_object()
        slug = self.kwargs.get(self.slug_url_kwarg, None)
        match = resolve(request.path_info)
        url_parts = match.namespaces
        url_parts.append(match.url_name)
        current_urlpattern = ":".join(url_parts)

        # Figure out what the slug is supposed to be.
        if hasattr(obj, "get_canonical_slug"):
            canonical_slug = obj.get_canonical_slug()
        else:
            canonical_slug = self.get_canonical_slug()

        # If there's a discrepancy between the slug in the url and the
        # canonical slug, redirect to the canonical slug.
        if canonical_slug != slug:
            params = {
                self.pk_url_kwarg: obj.pk,
                self.slug_url_kwarg: canonical_slug,
                "permanent": True,
            }
            return redirect(current_urlpattern, **params)

        return super().dispatch(request, *args, **kwargs)

    def get_canonical_slug(self):
        """
        Override this method to customize what slug should be considered
        canonical.
        Alternatively, define the get_canonical_slug method on this view's
        object class. In that case, this method will never be called.
        """
        return self.get_object().slug



from django import template

from ephios.extra.crispy import as_crispy_field

register = template.Library()


@register.simple_tag(name="crispy_field")
def crispy_field(field, label_class="", field_class="", wrapper_class=""):
    return as_crispy_field(
        field, label_class=label_class, field_class=field_class, wrapper_class=wrapper_class
    )



import datetime
import itertools

from django.db import models
from django.template.defaultfilters import yesno
from django.utils import formats
from django.utils.translation import gettext_lazy as _

# This file includes some of the itertools recipes from https://docs.python.org/3/library/itertools.html


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def partition(pred, iterable):
    "Use a predicate to partition entries into false entries and true entries"
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    return itertools.filterfalse(pred, t1), filter(pred, t2)


def format_anything(value):
    """Return some inbuilt types in a human readable way."""
    if isinstance(value, bool):
        return yesno(value)
    if isinstance(value, datetime.datetime):
        return formats.date_format(value, format="SHORT_DATETIME_FORMAT")
    if isinstance(value, models.QuerySet):
        return ", ".join(str(e) for e in value)
    if value is None:
        return _("None")
    return str(value)



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ephios.modellogging.json import LogJSONDecoder, LogJSONEncoder
from ephios.modellogging.recorders import (
    InstanceActionType,
    capitalize_first,
    recorder_types_by_slug,
)

# pylint: disable=protected-access


class LogEntry(models.Model):
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        related_name="logentries",
    )
    content_object_id = models.PositiveIntegerField(db_index=True)
    content_object = GenericForeignKey("content_type", "content_object_id")
    attached_to_object_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="associated_logentries"
    )
    attached_to_object_id = models.PositiveIntegerField(db_index=True)
    attached_to_object = GenericForeignKey("attached_to_object_type", "attached_to_object_id")
    datetime = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="logging_entries",
    )
    action_type = models.CharField(
        max_length=255, choices=[(value, value) for value in InstanceActionType]
    )
    request_id = models.CharField(max_length=36, null=True, blank=True)
    data = models.JSONField(default=dict, encoder=LogJSONEncoder, decoder=LogJSONDecoder)

    class Meta:
        ordering = ("-datetime", "-id")
        verbose_name = _("Log entry")
        verbose_name_plural = _("Log entries")

    @cached_property
    def records(self):
        recorder_types = recorder_types_by_slug(self.content_type.model_class())
        for recorder in self.data.values():
            if not isinstance(recorder, dict) or "slug" not in recorder:
                continue
            yield recorder_types[recorder["slug"]].deserialize(
                recorder["data"], self.content_type.model_class(), self.action_type
            )

    @property
    def content_object_classname(self):
        try:
            return capitalize_first(self.content_type.model_class()._meta.verbose_name)
        except AttributeError:
            return str(self.content_type)

    @property
    def content_object_or_str(self):
        try:
            return self.content_object
        except AttributeError:
            return self.data.get("__str__")

    def __str__(self):
        if self.content_object:
            return f"{self.action_type} {type(self.content_object)._meta.verbose_name} {str(self.content_object)}"
        return f"{self.action_type} {self.content_type.model} {self.content_object_or_str}"



from django import forms
from django.utils.translation import gettext_lazy as _

from ephios.core.models import AbstractParticipation
from ephios.core.signup.methods import ActionDisallowedError, BaseSignupMethod
from ephios.plugins.basesignup.signup.common import RenderParticipationPillsShiftStateMixin


class NoSelfserviceSignupMethod(RenderParticipationPillsShiftStateMixin, BaseSignupMethod):
    slug = "no_selfservice"
    verbose_name = _("No Signup (only disposition)")
    description = _("""This method allows no signup by users.""")
    uses_requested_state = False

    @property
    def configuration_form_class(self):
        class ConfigurationForm(super().configuration_form_class):
            no_selfservice_explanation = forms.CharField(
                label=_("Explanation"),
                required=False,
                initial="",
            )

        return ConfigurationForm

    def _configure_participation(
        self, participation: AbstractParticipation, **kwargs
    ) -> AbstractParticipation:
        raise TypeError(f"{self.__class__} does not support signup")

    @staticmethod
    def signup_is_disabled(method, participant):
        return ActionDisallowedError(
            method.configuration.no_selfservice_explanation
            or _("Signup for this shift is disabled.")
        )

    @property
    def _signup_checkers(self):
        return [
            self.signup_is_disabled,
        ]

    @property
    def _decline_checkers(self):
        return [
            self.signup_is_disabled,
        ]



from django.contrib import admin
from guardian.admin import GuardedModelAdmin

from ephios.core.models import (
    Consequence,
    Event,
    EventType,
    LocalParticipation,
    Notification,
    Qualification,
    QualificationCategory,
    QualificationGrant,
    Shift,
    WorkingHours,
)
from ephios.core.models.events import PlaceholderParticipation

admin.site.register(Qualification)
admin.site.register(QualificationGrant)
admin.site.register(QualificationCategory)
admin.site.register(WorkingHours)
admin.site.register(Consequence)

admin.site.register(Shift)
admin.site.register(Event, GuardedModelAdmin)
admin.site.register(EventType)
admin.site.register(LocalParticipation)
admin.site.register(PlaceholderParticipation)
admin.site.register(Notification)



from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _


def render_alert(content, alert_type=None, dismissible=True):
    """Render a Bootstrap alert."""
    button = ""
    if not alert_type:
        alert_type = "info"
    css_classes = ["alert", f"alert-{alert_type}"]
    if dismissible:
        css_classes.append("alert-dismissible")
        close = _("close")
        button = f'<button type="button" class="btn-close" data-bs-dismiss="alert" aria-label="{close}"></button>'
    return mark_safe(f"<div class='{' '.join(css_classes)}' role='alert'>{content}{button}</div>")



import datetime

from django import forms


class EndOfDayDateTimeField(forms.DateTimeField):
    """
    A datetime form field that when used with a date-only widget
    creates a datetime with the time at the end of the given day.
    """

    def to_python(self, value):
        result = super().to_python(value)
        if result is None:
            return result
        return datetime.datetime.max.replace(
            year=result.year,
            month=result.month,
            day=result.day,
        )
